import logging

from nicegui import ui

from ..pages import Pages, show_global_header
from ..session import get_cert_list, get_enddevice_list

_log = logging.getLogger(__name__)


@ui.page('/certs')
def show_certs():
    show_global_header(Pages.CERTS)
    device_list = get_enddevice_list()
    certs = get_cert_list()
    row_data = []
    for name, cert in certs.items():
        row = dict(name=f'{name}',
                   cert=f'<a href="/admin/download/cert/{name}">Cert File</a> &nbsp; <a href="/admin/download/key/{name}">Key File</a>',
                   path=f"{cert.get('path', '')}",
                   lFDI='')
        if cert.get('lFDI'):
            for ed in device_list.EndDevice:
                if ed.lFDI == cert['lFDI'].encode('utf-8'):
                    row['lFDI'] = ed.lFDI
                    break
            
        row_data.append(row)
    # for ed in device_list.EndDevice:
    #     row = dict(name=f'{ed.lFDI}',
    #             cert=f'<a href="/admin/cert/{ed.lFDI}">cert</a>')
    #     row_data.append(row)
        

    ui.aggrid({
        'columnDefs': [
            {'headerName': '', 'field': 'cert'},
            {'headerName': 'Path', 'field': 'path'},
            {'headerName': 'Name', 'field': 'name'},
            {'headerName': 'lFDI', 'field': 'lFDI'}
        ],
        'rowData': row_data,
    }, html_columns=[0])
    


import logging
from copy import deepcopy
from pprint import pformat
from typing import Any, Dict, List

from nicegui import ui

import ieee_2030_5.models as m

from ..pages import Pages, show_global_header
from ..session import get_enddevice_list, send_enddevice

_log = logging.getLogger(__name__)



enddevice_list: m.EndDeviceList = m.EndDeviceList()

default_enddevice: m.EndDevice = m.EndDevice(enabled=True, postRate=900)
current_enddevice: m.EndDevice = m.EndDevice(enabled=True, postRate=900)

class CurrentValueDict(dict):
    def __setitem__(self, __key: Any, __value: Any) -> None:
        return super().__setitem__(__key, __value)
    
    def __getitem__(self, __key: Any) -> Any:
        return super().__getitem__(__key)

current_enddevice = CurrentValueDict(current_enddevice.__dict__)


# Readonly list of controls for the program.
control_list = m.DERControlList()


@ui.refreshable
def render_select():
    
    end_device_hrefs = [p.href for p in enddevice_list.EndDevice]
    end_device_hrefs.insert(0, "NEW")
    
    with ui.row():
        with ui.column():
            if not current_enddevice.href:
                ui.select(end_device_hrefs, label="Programs", value=end_device_hrefs[0],
                          on_change=lambda e: change_select(e.value)).classes("w-64")
            else:
                ui.select(end_device_hrefs, label="Programs", value=current_enddevice.href,
                          on_change=lambda e: change_select(e.value)).classes("w-64")

def change_select(selected_value: str):
    global current_enddevice
    
    if selected_value in (None, "NEW"):
        _log.info(selected_value)
        current_enddevice = deepcopy(default_enddevice)
    else:
        current_enddevice = next(filter(lambda x: x.href == selected_value, enddevice_list.EndDevice))
        _log.info(selected_value)
        
    _log.debug(pformat(current_enddevice.__dict__))
    render_enddevice_form.refresh()

@ui.refreshable
def render_enddevice_form():
    
    with ui.row():
        with ui.column().classes("pr-20"):
            deviceCategory = ui.select({v.value: v.name for i, v in enumerate(m.DeviceCategoryType)}, 
                                       label="Device Category").classes("w-96") \
                .bind_value(current_enddevice, "deviceCategory", 
                            #forward=lambda x: x,
                            backward=lambda x: m.DeviceCategoryType(current_enddevice.deviceCategory).value)
                #.bind_value(current_enddevice, "deviceCategory")
            enabled = ui.checkbox("Enabled").bind_value(current_enddevice, "enabled")
            postRate = ui.number("postRate").bind_value(current_enddevice, "postRate")
        
        with ui.column():
            lfdi = ui.label(f"lFDI: {current_enddevice.lFDI}").bind_visibility(current_enddevice, 
                                                                        forward=lambda x: current_enddevice.href is None,
                                                                        backward=lambda x: current_enddevice.href is not None)
            
            sfdi = ui.label(f"sFDI: {current_enddevice.sFDI}").bind_visibility(current_enddevice, 
                                                                        forward=lambda x: current_enddevice.href is None,
                                                                        backward=lambda x: current_enddevice.href is not None)
            
        # TODO deal with function set url here
        # ui.select(names, multiple=True, value=names[:2], label='with chips') \
        #   .classes('w-64').props('use-chips')
            
def revert_changes():
    pass

def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global enddevice_list
    
    if append_new := current_enddevice.href is None:
        ...
        
    enddevice = send_enddevice(current_enddevice)
    
    assert isinstance(enddevice, m.EndDevice)
    
    if append_new:
        enddevice_list.EndDevice.append(enddevice)
    
    render_select.refresh()
    # global curve_data_points, current_curve
    # current_curve.CurveData = curve_data_points
    
    # if append_new_curve := current_curve.href is None:
    #     ...
    
    # # POST or PUT the curve to the server     
    # dc = send_control(current_curve)
    
    # assert isinstance(dc, m.DERCurve)
    
    # # update the variables for state
    # curve_data_points = dc.CurveData
    # current_curve = dc
    
    # # IF its post then we append to the curve list
    # if append_new_curve:
    #     curve_list.DERCurve.append(dc)
    
    # render_select.refresh()
    
    ui.notify("End Device saved successfully")
    
        
@ui.page(Pages.ENDDEVICES.value.uri)
def show_enddevices():
    global enddevice_list
    
    response = get_enddevice_list()
    if response:
        enddevice_list = response
        
        
    show_global_header(Pages.CONTROLS)
    render_select()
    render_enddevice_form()
    
    ui.separator()
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)


class AlreadyExistsError(Exception):
    """
    The AlreadyExistsError exception should be raise any time an object is attempted to be inserted
    into a collection where replacement is not allowed.
    """
    pass


class NotFoundError(Exception):
    pass



from ieee_2030_5_gui.views.server_control import server_control_view
from ieee_2030_5_gui.views.show_resource_tree import show_resource_tree_view
from ieee_2030_5_gui.views.home import home_view
from pydantic import BaseModel, ConfigDict
from typing import Optional, Callable
import flet as ft

class _2030_5View(BaseModel):
    #model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)
    route: str
    instance: object

class _2030_5Views(BaseModel):
    home: _2030_5View = _2030_5View(route="/", instance=home_view)
    server_control: _2030_5View = _2030_5View(route="/server_control", instance=server_control_view)
    show_tree: _2030_5View = _2030_5View(route="/show_tree", instance=show_resource_tree_view)

    def __getitem__(self, item):
        return getattr(self, item)

    def __iter__(self):
        for page in self.__dict__.values():
            yield page

AppViews = _2030_5Views()



import logging
from typing import Optional

import werkzeug.exceptions
from flask import Response, request

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.data.indexer import add_href, get_href
from ieee_2030_5.models import Registration
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.types_ import Lfdi
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass

_log = logging.getLogger(__name__)


class EDevRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def put(self) -> Response:
        parsed = hrefs.EdevHref.parse(request.path)

        mysubobj = xml_to_dataclass(request.data.decode('utf-8'))

        if get_href(request.path):
            response_status = 204
        else:
            response_status = 201

        add_href(request.path, mysubobj)

        return Response(status=response_status)

    def post(self, path: Optional[str] = None) -> Response:
        """
        Handle post request to /edev
        
        The expectation is that data will be an xml object like the following:
        
            <EndDevice xmlns="urn:ieee:std:2030.5:ns">
                <sFDI>231589308001</sFDI>
                <changedTime>0</changedTime>
            </EndDevice>
        
        Args:
            path: 

        Returns:

        """
        # request.data should have xml data.
        if not request.data:
            raise werkzeug.exceptions.Forbidden()

        ed: m.EndDevice = xml_to_dataclass(request.data.decode('utf-8'))

        if not isinstance(ed, m.EndDevice):
            raise werkzeug.exceptions.Forbidden()

        # This is what we should be using to get the device id of the registered end device.
        device_id = self.tls_repo.find_device_id_from_sfdi(ed.sFDI)
        ed.lFDI = self.tls_repo.lfdi(device_id)
        if end_device := adpt.EndDeviceAdapter.fetch_by_lfdi(ed.lfdi):
            status = 200
            ed_href = end_device.href
        else:
            if not ed.href:
                ed = adpt.EndDeviceAdapter.store(device_id, ed)

            ed_href = ed.href
            status = 201

        return Response(status=status, headers={'Location': ed_href})

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /edev
            /edev/0
            /edev/0/di
            /edev/0/rg
            /edev/0/der

        """
        # TODO start implementing these.
        start = int(request.args.get("s", 0))
        limit = int(request.args.get("l", 1))
        after = int(request.args.get("a", 0))

        edev_href = hrefs.HrefParser(request.path)

        ed = adpt.EndDeviceAdapter.fetch_by_property('lFDI', self.lfdi)

        # /edev_0_dstat
        if request.path == ed.DERListLink.href:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
        elif request.path == ed.LogEventListLink.href:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
        elif request.path == ed.FunctionSetAssignmentsListLink.href:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
        elif edev_href.count() > 2:
            retval = get_href(request.path)
        elif not edev_href.has_index():
            retval = m.EndDeviceList(href=request.path, all=1, results=1, EndDevice=[ed])
        else:
            retval = get_href(request.path)

        # if adpt.ListAdapter.has_list(request.path):
        #     retval = adpt.ListAdapter.get_resource_list(request.path)

        return self.build_response_from_dataclass(retval)


class SDevRequests(RequestOp):
    """
    SelfDevice is an alias for the end device of a client.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /sdev

        """
        end_device = self._end_devices.get_end_device_list(self.lfdi).EndDevice[0]
        return self.build_response_from_dataclass(end_device)


class FSARequests(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self):
        """ Retrieve a FSA or Program List
        """
        start = int(request.args.get("s", 0))
        limit = int(request.args.get("l", 0))
        after = int(request.args.get("a", 0))

        fsa_href = hrefs.fsa_parse(request.path)

        if fsa_href.fsa_index == hrefs.NO_INDEX:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
            # retval = adpt.FunctionSetAssignmentsAdapter.fetch_all(m.FunctionSetAssignmentsList(),
            #                                                       "FunctionSetAssignments")
        elif fsa_href.fsa_sub == hrefs.FSASubType.DERProgram.value:
            retval = adpt.ListAdapter.get_resource_list(request.path, start, after, limit)
            # fsa = adpt.FunctionSetAssignmentsAdapter.fetch(fsa_href.fsa_index)
            # retval = adpt.FunctionSetAssignmentsAdapter.fetch_children(
            #     fsa, "fsa", m.DERProgramList())
            # # retval = FSAAdapter.fetch_children_list_container(fsa_href.fsa_index, m.DERProgram, m.DERProgramList(href="/derp"), "DERProgram")
        # else:
        #     retval = adpt.FunctionSetAssignmentsAdapter.fetch(fsa_href.fsa_index)

        return self.build_response_from_dataclass(retval)


import sys
import time
from pprint import pprint
from typing import Any, Optional

import cimlab.loaders.gridappsd as gridappsd_loader
from cimlab.data_profile import CIM_PROFILE
from cimlab.loaders import ConnectionParameters, Parameter
from cimlab.loaders.gridappsd import GridappsdConnection
from cimlab.models import DistributedModel
from gridappsd import GridAPPSD

gridappsd_loader.set_cim_profile(CIM_PROFILE.RC4_2021.value)

cim = gridappsd_loader.cim

inverter_list = []
debug_output = False


def do_print(data: Any, *args):
    if debug_output:
        print(data, *args)


# sort PowerElectronicsUnits
def get_inverter_buses(network_area):
    if cim.PowerElectronicsConnection in network_area.typed_catalog:
        network_area.get_all_attributes(cim.PowerElectronicsConnection)
        network_area.get_all_attributes(cim.PowerElectronicsConnectionPhase)
        network_area.get_all_attributes(cim.Terminal)
        network_area.get_all_attributes(cim.Analog)

        do_print('\n \n EXAMPLE 6: GET ALL INVERTER PHASES AND BUSES')
        for pec in network_area.typed_catalog[cim.PowerElectronicsConnection].values():
            do_print('\n name: ', pec.name, pec.mRID)
            inverter_list.append((pec.name, pec.mRID, pec.p, pec.q))
            do_print('p = ', pec.p, 'q = ', pec.q)
            node1 = pec.Terminals[0].ConnectivityNode
            do_print('bus: ', node1.name, node1.mRID)
            for pec_phs in pec.PowerElectronicsConnectionPhases:
                do_print('phase ', pec_phs.phase, ': ', pec_phs.mRID)

            for meas in pec.Measurements:
                do_print('Measurement: ', meas.name, meas.mRID)
                do_print('type:', meas.measurementType, 'phases:', meas.phases)


def _main():
    topic = "goss.gridappsd.request.data.topology"
    #feeder_mrid = "_C07972A7-600D-4AA5-B254-4CAA4263560E"    # Ochre 13-node
    feeder_mrid = "_49AD8E07-3BF9-A4E2-CB8F-C3722F837B62"    # 13-node
    # feeder_mrid = "_E407CBB6-8C8D-9BC9-589C-AB83FBF0826D"    # 123-node
    message = {"requestType": "GET_SWITCH_AREAS", "modelID": feeder_mrid, "resultFormat": "JSON"}

    gapps = GridAPPSD(username="system", password="manager")
    t1 = time.perf_counter()
    topology_response = gapps.get_response(topic=topic, message=message, timeout=30)
    t2 = time.perf_counter()
    do_print(f"Retriving topology took: {t2-t1:0}s")

    # TODO setup environment and/or pass in parameters for env.
    gapps_conn = GridappsdConnection(ConnectionParameters())
    gapps_conn.connect()
    feeder = cim.Feeder(mRID=feeder_mrid)
    network = DistributedModel(connection=gapps_conn,
                               feeder=feeder,
                               topology=topology_response['feeders'])

    t1 = time.perf_counter()
    for switch_area in network.switch_areas:
        get_inverter_buses(switch_area)
        for secondary_area in switch_area.secondary_areas:
            get_inverter_buses(secondary_area)
    t2 = time.perf_counter()
    do_print(f"Retriving inverters: {t2-t1:0}s")

    # do_print(network.typed_catalog)
    for t in network.typed_catalog:
        do_print(t.__name__)
    # do_print(network.typed_catalog.get(cim.PowerElectronicsConnection))
    for name, mrid, p, q in inverter_list:
        equip_type = "Solar Panel" if int(p) > 0 else "Battery"
        print(f"name: {name}, equipment: {equip_type}, mrid: {mrid}, p: {p}, q: {q}")
    # pprint(inverter_list)


if __name__ == '__main__':
    _main()


import logging
from copy import deepcopy
from typing import Dict, List

from nicegui import ui

import ieee_2030_5.models as m

from ..pages import Pages, show_global_header
from ..session import get_der_list, send_der

_log = logging.getLogger(__name__)

page: Pages = Pages.DER

der_list: m.DERList = m.DERList()

current_der: m.DER = m.DER()

#program_list: m.DERProgramList() = get_program_list()


ui.refreshable
def render_select():
    
    der_selection = [p.description for p in der_list.DER]
    der_selection.insert(0, "NEW")
    
    with ui.row():
        with ui.column():
            if not current_der.description:
                ui.select(der_selection, label="DER", value=der_selection[0]).classes("w-64")
            else:
                ui.select(der_selection, label="DER", value=current_der.description).classes("w-64")
                
            if current_der.href:
                ui.label(f"Href: {current_der.href}")



@ui.refreshable
def render_form():
    with ui.row():
        ui.input("description").classes("w-64") \
            .bind_value(current_der, "description")
    ui.separator()
    
    with ui.row():
        with ui.column():
            pass
                
    with ui.row():
        with ui.column():
            pass

            
            
def revert_changes():
    pass

def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global der_base, der_default
    
    if append_new := current_der.href is None:
        ...
        
    der = send_der(current_der)
    
    assert isinstance(der, m.DER)
    
    if append_new:
        der_list.DER.append(der)
    
    render_select.refresh()
    
    ui.notify(f"{page.value.title} saved successfully")
    
        
@ui.page(page.value.uri)
def show_der():
    global der_list
    
    response = get_der_list()
    if response:
        der_list = response
        
        
    show_global_header(Pages.DER)
    render_select()
    render_form()
    
    ui.separator()
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)


import logging
from copy import deepcopy
from typing import Dict, List

from nicegui import ui

import ieee_2030_5.models as m

from ..pages import Pages, show_global_header
from ..session import get_control_list, get_program_list, send_program

_log = logging.getLogger(__name__)

page: Pages = Pages.PROGRAMS

programs: m.DERProgramList = m.DERProgramList()


current_primacy_type: m.PrimacyType = m.PrimacyType.InHomeManagementSystem
current_program: m.DERProgram = m.DERProgram(primacy=current_primacy_type.value,
                                             DefaultDERControlLink=m.DefaultDERControlLink(),
                                             ActiveDERControlListLink=m.ActiveDERControlListLink(),
                                             DERControlListLink=m.DERControlListLink())

# Readonly list of controls for the program.
control_list: m.DERControlList = get_control_list()

ui.refreshable
def render_select():
    
    der_programs = [p.description for p in programs.DERProgram]
    der_programs.insert(0, "NEW")
    
    with ui.row():
        with ui.column():
            if not current_program.description:
                ui.select(der_programs, label="Programs", value=der_programs[0]).classes("w-64")
            else:
                ui.select(der_programs, label="Programs", value=current_program.description).classes("w-64")
                
            if current_program.href:
                ui.label(f"Href: {current_program.href}")

def render_primacy_select():
    global current_primacy_type
    
    with ui.row():
        with ui.column():
            current_primacy_type = ui.select({v.value: v.name for i, v in enumerate(m.PrimacyType)}, label="Primacy").classes("w-96") \
                .bind_value(current_program, "primacy")
                
def render_default_control_select():
    der_controls = { p.href: p.description for p in control_list.DERControl}
    
    with ui.row():
        with ui.column():
            ui.select(der_controls, label="Default Control", value=current_program.DefaultDERControlLink.href).classes("w-64")
    
def change_control(new_control: str):
    pass    

@ui.refreshable
def render_form():
    if current_program.href:
        with ui.row():
            with ui.column():
                ui.label(f"Href: {current_program.href}")
                ui.label(f"MRID: {current_program.mRID}")
                ui.label(f"version: {current_program.version}")
    
    with ui.row():
        ui.input("description").classes("w-64") \
            .bind_value(current_program, "description")
    
    ui.separator()
    
    render_primacy_select()
    render_default_control_select()
                
    with ui.row():
        with ui.column():
            pass
            
            
            
            
            
def revert_changes():
    pass

def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global der_base, der_default
    
    if append_new_program := current_program.href is None:
        ...
        
    program = send_program(current_program)
    
    assert isinstance(program, m.DERProgram)
    
    if append_new_program:
        program_list.DERProgram.append(program)
    
    render_select.refresh()
    
    ui.notify(f"{page.value.title} saved successfully")
    
        
@ui.page(page.value.uri)
def show_programs():
    global program_list
    
    program_response = get_program_list()
    if program_response:
        program_list = program_response

    show_global_header(Pages.CONTROLS)
    render_select()
    render_form()
    
    ui.separator()
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)


from fastapi import Request
from fastapi.middleware.wsgi import WSGIMiddleware
import flet as ft
from flet.fastapi import FastAPI
import os
from pathlib import Path
#from ieee_2030_5.server
#from asyncio import asynccontextmanager


# Hook flet into FastAPI using flet.fastapi module
# @asynccontextmanager
# async def lifespan(app: FastAPI):
#     await flet_fastapi.app_manager.start()
#     yield
#     await flet_fastapi.app_manager.shutdown()

app = FastAPI()

os.environ["FLET_SECRET_KEY"] = "secret"

@app.get("/")
def read_root(request: Request):
    return {"Hello": "World"}


def gui_main(page: ft.Page):
    from ieee_2030_5_gui.views import AppViews as views
    # import uudex_web.data as data
    # from uudex_web.views import UUDEXViews as uudex_views

    try:
        page.horizontal_alignment = ft.CrossAxisAlignment.CENTER
        page.title = "IEEE 2030.5 Admin GUI"

        def route_change(e: ft.RouteChangeEvent):
            page.views.clear()

            for v in views:
                if v.route == e.route:
                    page.views.append(v.instance(page=page))
                    break
            if not page.views:
                raise ValueError(f"The route {e.route} was not defined.")

            page.update()

        def view_pop(view):
            page.views.pop()
            top_view = page.views[-1]
            page.go(top_view.route)

        page.on_route_change = route_change
        page.on_view_pop = view_pop
        page.go(page.route)

    except Exception as e:
        _log.exception(e)


# Mount the flet app to the fastapi app.  More than one endpoint can be mounted to the same fastapi app.
app.mount(
    "/gui",
    ft.app(gui_main, export_asgi_app=True,
                     web_renderer=ft.WebRenderer.AUTO,
                     upload_dir=Path("~/.2030_5_uploads").expanduser().as_posix()))


def _main():
    from argparse import ArgumentParser
    import os

    parser = ArgumentParser()

    parser.add_argument("config", help="Path to config server configuration file.")
    parser.add_argument("--reset-certs", action="store_true", help="Reset the certificates.")

    opts = parser.parse_args()

    os.environ["CONFIG_FILE"] = opts.config
    os.environ["RESET_CERTS"] = str(opts.reset_certs)

    import uvicorn

    uvicorn.run(app, host="localhost", port=8000)

if __name__ == "__main__":
    from argparse import ArgumentParser
    import os

    parser = ArgumentParser()

    parser.add_argument("config", help="Path to config server configuration file.")
    parser.add_argument("--reset-certs", action="store_true", help="Reset the certificates.")

    opts = parser.parse_args()

    os.environ["CONFIG_FILE"] = opts.config
    os.environ["RESET_CERTS"] = str(opts.reset_certs)



    import uvicorn

    uvicorn.run(app, host="localhost", port=8000)

# from __future__ import annotations

# #import io
# import logging
# import os
# import sys

# #from dotenv import load_dotenv
# #from fastapi import FastAPI
# #from nicegui import ui


# logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
# _log = logging.getLogger(__name__)

# # @ui.page("/")
# # def run_page():
# #     ui.markdown("""Ignoring this""")

# # ui.run()

# # app = FastAPI()

# # __all__ = [
# #     "run_gui"
# # ]
# def setup_environment():
#     """Creates the os.environ for the application.

#     This is done using the .env file in same directory as this file or from
#     the command line parameters or from the environmental variables already existing.
#     """
#     from pathlib import Path

#     PREFIX = "2030_5"

#     # Allow custmization of the 2030_5_ENV_PATH
#     env_path = Path(os.environ.get(f"{PREFIX}_ENV_PATH",
#                                     str(Path(__file__).parent / ".env")))


#     load_dotenv(dotenv_path=env_path.as_posix())

#     required_env = [
#         f"{PREFIX}_HOST",
#         f"{PREFIX}_PORT",
#         f"{PREFIX}_CLIENT_KEY",
#         f"{PREFIX}_CLIENT_CERT",
#         f"{PREFIX}_CA_CERT"
#     ]

#     for x in required_env:
#         if os.getenv(x) is None or os.getenv(x).strip() == "":
#             sys.stderr.write(f'Missing {x} in .env file\n')
#             sys.exit(1)

#     paths = [
#         f"{PREFIX}_CLIENT_KEY",
#         f"{PREFIX}_CLIENT_CERT",
#         f"{PREFIX}_CA_CERT"
#     ]

#     for x in paths:
#         os.environ[x] = Path(os.getenv(x)).expanduser().as_posix()

# def initialize_app():

#     show_global_header(Pages.HOME)

#     ui.markdown("""

# # Welcome to the IEEE 2030.5 Admin GUI

# This is a work in progress""")

# if __name__ in {"__main__", "__mp_main__"}:
#     # out_file = open(sys.stdout.fileno(), 'wb', 0)
#     # sys.stdout = io.TextIOWrapper(out_file, write_through=True)

#     # Must be done before importing the rest of the application.
#     setup_environment()

#     from ieee_2030_5_gui.pages import Pages, show_global_header

#     logging.basicConfig(level=logging.DEBUG)

#     initialize_app()

#     # excludes = ['data_store/**',
#     #             'ieee_2030_5/adapters/**',
#     #             'ieee_2030_5/client',
#     #             'ieee_2030_5/data]
#     includes = ['ieee_2030_5_gui/**',]

#     ui.run(show=False, uvicorn_reload_includes=",".join(includes))

#     # from pathlib import Path

#     # env_path = Path(__file__).parent / ".env"

#     # load_dotenv(dotenv_path=env_path.as_posix())

#     # PREFIX = "2030_5"
#     # required_env = [
#     #     f"{PREFIX}_HOST",
#     #     f"{PREFIX}_PORT",
#     #     f"{PREFIX}_CLIENT_KEY",
#     #     f"{PREFIX}_CLIENT_CERT",
#     #     f"{PREFIX}_CA_CERT"
#     # ]

#     # for x in required_env:
#     #     if os.getenv(x) is None or os.getenv(x).strip() == "":
#     #         sys.stderr.write(f'Missing {x} in .env file\n')
#     #         sys.exit(1)

#     # logging.basicConfig(level=logging.DEBUG)

#     # initialize_app()


#     # ui.run(show=False)





# # ui.html("""
# #     <table>
# #         <tr>
# #             <td>1</td>
# #             <td>2</td>
# #         </tr>
# #     </table>
# # """)


# # columns = [
# #     {'name': 'name', 'label': 'Name', 'field': 'name', 'required': True},
# #     {'name': 'age', 'label': 'Age', 'field': 'age', 'sortable': True},
# # ]
# # rows = [
# #     {'id': 0, 'name': 'Alice', 'age': 18},
# #     {'id': 1, 'name': 'Bob', 'age': 21},
# #     {'id': 2, 'name': 'Lionel', 'age': 19},
# #     {'id': 3, 'name': 'Michael', 'age': 32},
# #     {'id': 4, 'name': 'Julie', 'age': 12},
# #     {'id': 5, 'name': 'Livia', 'age': 25},
# #     {'id': 6, 'name': 'Carol'},
# # ]

# # with ui.table(title='My Team', columns=columns, rows=rows, selection='multiple', pagination=10).classes('w-96') as table:
# #     with table.add_slot('top-right'):
# #         with ui.input(placeholder='Search').props('type=search').bind_value(table, 'filter').add_slot('append'):
# #             ui.icon('search')
# #     with table.add_slot('body'):
# #         with table.row() as row:
# #             with table.cell() as cell:
# #                 pprint(row.props)
# #                 ui.link('Download', f'https://nicegui.io?{row.props.id}')
# #             with table.cell():
# #                 ui.link('Download', 'https://nicegui.io')
# #             with table.cell():
# #                 ui.link('Download', 'https://nicegui.io')
# #     with table.add_slot('bottom-row'):
# #         with table.row():
# #             with table.cell():
# #                 ui.button(on_click=lambda: (
# #                     table.add_rows({'id': time.time(), 'name': new_name.value, 'age': new_age.value}),
# #                     new_name.set_value(None),
# #                     new_age.set_value(None),
# #                 ), icon='add').props('flat fab-mini')
# #             with table.cell():
# #                 new_name = ui.input('Name')
# #             with table.cell():
# #                 new_age = ui.number('Age')

# # ui.label().bind_text_from(table, 'selected', lambda val: f'Current selection: {val}')
# # ui.button('Remove', on_click=lambda: table.remove_rows(*table.selected)) \
# #     .bind_visibility_from(table, 'selected', backward=lambda val: bool(val))

#     #ui.run(show=False)



"""
This module handles MirrorUsagePoint and UsagePoint constructs for a server.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

from flask import Response, request
from werkzeug.exceptions import BadRequest

import ieee_2030_5.adapters as adpt
import ieee_2030_5.models as m
from ieee_2030_5 import hrefs
from ieee_2030_5.data.indexer import get_href
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.server.uuid_handler import UUIDHandler
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass


class Error(Exception):
    pass


@dataclass
class ResponseStatus:
    location: str
    status: str


class UsagePointsContainer:
    __upt__: Dict[bytes, m.UsagePoint] = {}
    __mup__: Dict[bytes, m.MirrorUsagePoint] = {}
    __sorted_mrid__: List = []
    __mup_href__: Dict[str, m.MirrorUsagePoint] = {}
    __mup_readings__: Dict[bytes, m.MirrorReadingSet] = {}

    def get_mup_list(self,
                     start: Optional[int] = None,
                     after: Optional[int] = None,
                     length: Optional[int] = None) -> m.MirrorUsagePointList:

        if start is not None and after is not None:
            # after takes precedence
            index = after + 1 + start
        elif start is not None:
            index = start
        elif after is not None:
            index = after + 1
        else:
            index = 0

        mup_list = m.MirrorUsagePointList(all=len(self.__sorted_mrid__))

        if length is None:
            for x in self.__sorted_mrid__[index:len(self.__sorted_mrid__)]:
                mup_list.MirrorUsagePoint.append(self.__mup__[x])
        elif length - index < len(self.__sorted_mrid__):
            for x in self.__sorted_mrid__[index:index + length]:
                mup_list.MirrorUsagePoint.append(self.__mup__[x])
        else:
            for x in self.__sorted_mrid__[index:]:
                mup_list.MirrorUsagePoint.append(self.__mup__[x])

        mup_list.results = len(mup_list.MirrorUsagePoint)

        return mup_list

    def get_upt_list(self) -> m.UsagePointList:
        pass

    def get_mup_href(self, href: str) -> Optional[m.MirrorUsagePoint]:
        return self.__mup_href__.get(href)

    def create_reading_set(self, mup_href: str, mrs: m.MirrorReadingSet) -> ResponseStatus | Error:
        mup = point_container.get_mup_href(mup_href)
        if mup is None:
            return Error(f"No MirrorUsagePoint found at posted: {mup_href}")

        #     point_container.create_reading_set(mup_href=)

    def create_update_mup(self, mup: m.MirrorUsagePoint) -> ResponseStatus | Error:
        required = ["mRID", "deviceLFDI"]
        invalid = []
        for p in required:
            if not getattr(mup, p):
                invalid.append(f"{p} not specified for MirrorUsagePoint")
        # Error out top level required data
        if invalid:
            return Error("\n".join(invalid))

        required_meter_reading = ["mRID", "ReadingType"]
        for indx, reading in enumerate(mup.MirrorMeterReading):
            for p in required_meter_reading:
                if not getattr(reading, p):
                    invalid.append(f"Reading {indx} missing {p}")
        # Error out if meter reading invalid.
        if invalid:
            return Error("\n".join(invalid))

        created = self.__mup__.get(mup.mRID) is None
        self.__mup__[mup.mRID] = mup
        self.__sorted_mrid__.append(mup.mRID)
        # Sort mrids descending
        self.__sorted_mrid__ = sorted(self.__sorted_mrid__, reverse=True)

        item_in_list = 0
        for indx, mrid in enumerate(self.__sorted_mrid__):
            if mrid == mup.mRID:
                item_in_list = indx
                break
        mup.href = hrefs.build_link(hrefs.DEFAULT_MUP_ROOT, item_in_list)
        return ResponseStatus(mup.href, '201 Created' if created else '204 Updated')


point_container = UsagePointsContainer()


class UTP(RequestOp):
    __next_index__: int = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        pass

    def create(self, mrid: Optional[str] = None) -> m.UsagePoint:
        up = m.UsagePoint(href=f"{hrefs.build_link(hrefs.upt, UTP.__next_index__)}", mRID=mrid)
        UTP.__next_index__ += 1


class MUP(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self, path) -> Response:
        pth_info = request.environ['PATH_INFO']

        if not pth_info.startswith(hrefs.DEFAULT_MUP_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        ret_value = get_href(pth_info)
        if pth_info == hrefs.DEFAULT_MUP_ROOT:
            # TODO add list ability here.
            ret_value = adpt.MirrorUsagePointAdapter.fetch_all()
            # # Getting list of elements
            # start = request.args.get("s")
            # after = request.args.get("a")
            # length = request.args.get("l")
            # ret_value = adpt.MirrorUsagePointAdapter.get_list(start, after, length)
            #ret_value = point_container.get_mup_list(start, after, length)
        else:
            ret_value = get_href(pth_info)
            #ret_value = point_container.get_mup_href(pth_info)
        if ret_value:
            return self.build_response_from_dataclass(ret_value)

        return Response("Not Found", status=404)

    def post(self, path) -> Response:
        xml = request.data.decode('utf-8')
        data = xml_to_dataclass(request.data.decode('utf-8'))
        data_type = type(data)
        if data_type not in (m.MirrorUsagePoint, m.MirrorReadingSet, m.MirrorMeterReading):
            raise BadRequest()

        pth_info = request.path
        pths = pth_info.split("/")
        if len(pths) == 1 and data_type is not m.MirrorUsagePoint:
            # Check to make sure not a new mRID
            raise BadRequest("Must post MirrorUsagePoint to top level only")

        # Creating a new mup
        if data_type == m.MirrorUsagePoint:
            result = adpt.MirrorUsagePointAdapter.create(mup=data)
            if result:
                result = ResponseStatus(data.href, "201")
            else:
                result = ResponseStatus(data.href, "200")

        else:
            result = adpt.MirrorUsagePointAdapter.create_reading(href=pth_info, data=data)
            # MirrorReadingSet
            #result = point_container.create_reading_set(mup_href=pth_info, mrs=data)

        if isinstance(result, Error):
            return Response(result.args[1], status=500)
        # Note response to the post is different due to added endpoint.
        return Response(headers={'Location': result.location}, status=result.status)



from __future__ import annotations

import os
from dataclasses import asdict
import json
import logging

import re
from threading import Timer

ENABLED = True
try:
    from attrs import define, field
    import gridappsd.topics as topics
    from ieee_2030_5.types_ import Lfdi
    import ieee_2030_5.models as m
    from gridappsd import GridAPPSD
    from gridappsd.field_interface.interfaces import FieldMessageBus
    from cimgraph.data_profile import CIM_PROFILE
    from gridappsd.field_interface.agents.agents import GridAPPSDMessageBus
    import cimgraph.data_profile.rc4_2021 as cim
    from cimgraph.models import FeederModel
    from cimgraph.databases.gridappsd import GridappsdConnection
    from cimgraph.databases import ConnectionParameters

except ImportError:
    ENABLED = False

if ENABLED:
    _log = logging.getLogger("ieee_2030_5.gridappsd.adapter")
    from ieee_2030_5.certs import TLSRepository
    from ieee_2030_5.config import DeviceConfiguration, GridappsdConfiguration
    import ieee_2030_5.adapters as adpt

    @define
    class HouseLookup:
        mRID: str
        name: str
        lfdi: Lfdi


    class PublishTimer(Timer):
        # def __init__(self, interval: float, function, adapter: GridAPPSDAdapter):
        #     self.adapter = adapter
        #     super().__init__(interval=interval, function=function)
        def run(self):
            while not self.finished.wait(self.interval):
                self.function(*self.args, **self.kwargs)


    @define
    class GridAPPSDAdapter:
        gapps: GridAPPSD
        gridappsd_configuration: dict | GridappsdConfiguration
        tls: TLSRepository

        _publish_interval_seconds: int = 3
        _default_pin: str | None = None
        _inverters: list[HouseLookup] | None = None
        _model_dict_file: str | None = None
        _model_id: str | None = None
        _model_name: str | None = None
        _devices: list[DeviceConfiguration] | None = None
        _power_electronic_connections: list[cim.PowerElectronicsConnection] | None = None
        _timer: PublishTimer | None = None
        __field_bus_connection__: FieldMessageBus | None = None

        def get_message_bus(self) -> FieldMessageBus:
            if self.__field_bus_connection__ is None:
                # TODO Use factory class here!
                self.__field_bus_connection__ = GridAPPSDMessageBus(self.gridappsd_configuration.field_bus_def)
                # TODO Hack to make sure the gridappsd is actually able to connect.
                self.__field_bus_connection__.gridappsd_obj = GridAPPSD(username=self.gridappsd_configuration.username,
                                                                        password=self.gridappsd_configuration.password)
                # TODO Use the interface instead of this, however the gridappsdmessagebus doesn't implement it!
                assert self.__field_bus_connection__.gridappsd_obj.connected
            return self.__field_bus_connection__

        def use_houses_as_inverters(self) -> bool:
            return (self.gridappsd_configuration.house_named_inverters_regex is not None or
                    self.gridappsd_configuration.utility_named_inverters_regex is not None)

        def __attrs_post_init__(self):
            if self.gridappsd_configuration is not None and not isinstance(self.gridappsd_configuration,
                                                                           GridappsdConfiguration):
                self.gridappsd_configuration = GridappsdConfiguration(**self.gridappsd_configuration)

            if not self.gridappsd_configuration:
                raise ValueError("Missing GridAPPSD configuration, but it is required.")

            self._model_name = self.gridappsd_configuration.model_name
            self._default_pin = self.gridappsd_configuration.default_pin

            _log.debug("Creating timer now")
            self._timer = PublishTimer(self._publish_interval_seconds, self.publish_house_aggregates)
            self._timer.start()

        # power_electronic_connections: list[cim.PowerElectronicsConnection] = []

        def get_model_id_from_name(self) -> str:
            models = self.gapps.query_model_info()
            for m in models['data']['models']:
                if m['modelName'] == self._model_name:
                    return m['modelId']
            raise ValueError(f"Model {self._model_name} not found")

        def get_house_and_utility_inverters(self) -> list[HouseLookup]:
            """
            This function uses the GridAPPSD API to get the list of energy consumers.

            This method should only be called with the `house_named_inverters_regex` or `utility_named_inverters_regex`
            properties set on the `GridappsdConfiguration object.  If set then the function searches for energy
            consumers that match the regular expression and returns them as a list of HouseLookup objects.
            In the case of utility regular expression it will return 3 HouseLookup objects for each phase of the
            utility inverter.  The name of the phase (a b c, A B C, 1 2 3, etc) is determined by the
            response from the server in the querying of the model.

            :return: list of HouseLookup objects
            :rtype: list[HouseLookup]
            """

            if self._inverters is not None:
                return self._inverters

            self._inverters = []

            if self._model_dict_file is None:

                if self._model_id is None:
                    self._model_id = self.get_model_id_from_name()

                response = self.gapps.get_response(topic='goss.gridappsd.process.request.config',
                                                   message={"configurationType": "CIM Dictionary",
                                                            "parameters": {"model_id": f"{self._model_id}"}})

                # Should have returned only a single feeder
                feeder = response['data']['feeders'][0]
            else:

                with open(self.model_dict_file, 'r') as f:
                    feeder = json.load(f)['feeders'][0]

            re_houses = re.compile(self.gridappsd_configuration.house_named_inverters_regex)
            re_utility = re.compile(self.gridappsd_configuration.utility_named_inverters_regex)

            # Based upon the energyconsumers create matches to the houses and utilities
            # and add them to the list.
            for ec in feeder['energyconsumers']:
                if match_house := re.match(re_houses, ec['name']):
                    self._inverters.append(HouseLookup(mRID=ec['mRID'],
                                                       name=match_house.group(0),
                                                       lfdi=self.tls.lfdi(ec['mRID'])))
                elif match_utility := re.match(re_utility, ec['name']):
                    for phase in ec['phases']:
                        self._inverters.append(HouseLookup(mRID=ec['mRID'],
                                                           name=match_utility.group(0) + f"_{phase}",
                                                           lfdi=self.tls.lfdi(ec['mRID'])))
            return self._inverters

        def get_power_electronic_connections(self) -> list[cim.PowerElectronicsConnection]:
            if self._power_electronic_connections is not None:
                return self._power_electronic_connections

            self._power_electronic_connections = []

            models = self.gapps.query_model_info()
            for m in models['data']['models']:
                if m['modelName'] == self._model_name:
                    self._model_id = m['modelId']
                    break
            if not self._model_id:
                raise ValueError(f"Model {self._model_name} not found")

            cim_profile = CIM_PROFILE.RC4_2021.value
            iec = 7
            params = ConnectionParameters(cim_profile=cim_profile, iec61970_301=iec)

            conn = GridappsdConnection(params)
            conn.cim_profile = cim_profile
            feeder = cim.Feeder(mRID=self._model_id)

            network = FeederModel(connection=conn, container=feeder, distributed=False)

            network.get_all_edges(cim.PowerElectronicsConnection)

            self._power_electronic_connections = network.graph[cim.PowerElectronicsConnection].values()
            return self._power_electronic_connections

        def _build_device_configurations(self):
            self._devices = []
            if self.use_houses_as_inverters():
                for inv in self.get_house_and_utility_inverters():
                    dev = DeviceConfiguration(id=inv.mRID,
                                              pin=int(self._default_pin),
                                              lfdi=self.tls.lfdi(inv.mRID))
                    dev.ders = [dict(description=inv.name)]
                    self._devices.append(dev)
            else:
                for inv in self.get_power_electronic_connections():
                    dev = DeviceConfiguration(
                        id=inv.mRID,
                        pin=int(self._default_pin),
                        lfdi=self.tls.lfdi(inv.mRID)
                    )
                    dev.ders = [dict(description=inv.mRID)]
                    self._devices.append(dev)

        def get_device_configurations(self) -> list[DeviceConfiguration]:
            if not self._devices:
                self._build_device_configurations()
            return self._devices

        def get_message_for_bus(self) -> dict:
            import random
            msg = {}

            # TODO Get from list adapter for each house.
            # TODO This might not be the right way to do this
            # Filter for availability
            # for dev in self._devices:
            #     if inverter := next(filter(lambda x: x.lfdi == dev.lfdi, self._inverters):

            der_status_uris = adpt.ListAdapter.filter_single_dict(lambda k: k.endswith('ders'))

            for uri in der_status_uris:
                _log.debug(f"Testing uri: {uri}")

                meta_data = adpt.ListAdapter.get_single_meta_data(uri)
                status: m.DERStatus = adpt.ListAdapter.get_single(meta_data['uri'])

                inverter: HouseLookup | None = None

                _log.debug(f"Status is: {status}")
                if status:
                    _log.debug(f"Status found: {status}")
                    for x in self._devices:
                        if x.lfdi == meta_data['lfdi']:
                            inverter = x
                            _log.debug(f"Found inverter: {inverter}")
                            break

                    if inverter:

                        msg[inverter.mRID] = dict(mrid=inverter.mRID, name=inverter.name)

                        msg[inverter.mRID].update(asdict(status))


            # for inv in self._inverters:
            #     adpt.ListAdapter.get_single_meta_data(())
            #     try:
            #         # TODO we need to make sure we have the right mRID for the inverters this doesn't guarantee it.
            #         uri = next(der_status_uris)
            #         data = adpt.ListAdapter.get_single(uri)
            #         msg[inv.mRID] = dict(mrid=inv.mRID, name=inv.name)
            #         for k, v in data.__dict__.items():
            #             if v is not None:
            #                 msg[inv.mRID][k] = v
            #
            #     except StopIteration:
            #         pass

            return msg

        def create_2030_5_device_certificates_and_configurations(self) -> list[DeviceConfiguration]:

            self._devices = []
            if self.use_houses_as_inverters():
                for house in self.get_house_and_utility_inverters():
                    self.tls.create_cert(house.mRID)
            else:
                for inv in self.get_power_electronic_connections():
                    self.tls.create_cert(inv.mRID)
            self._build_device_configurations()
            return self._devices

        def publish_house_aggregates(self):
            from pprint import pformat

            mb = self.get_message_bus()

            if field_bus := self.gridappsd_configuration.field_bus_def.id:
                ...

            if simulation_id := os.environ.get("GRIDAPPSD_SIMULATION_ID"):
                pass
            if service_name := os.environ.get("GRIDAPPSD_SERVICE_NAME"):
                pass

            output_topic = topics.application_output_topic(application_id=service_name, simulation_id=simulation_id)
            # # TODO: the output topic goes to the field bus manager regardless of the message_bus_id for some reason.
            # output_topic = topics.field_output_topic(message_bus_id=field_bus)

            message = self.get_message_for_bus()

            _log.debug(f"Output: {output_topic}\n{pformat(message, 2)}")
            mb.send(topic=output_topic, message=message)



import logging
from copy import deepcopy
from typing import Dict, List

from nicegui import ui

import ieee_2030_5.models as m

from ..pages import Pages, show_global_header
from ..session import get_fsa_list, get_program_list, send_fsa

_log = logging.getLogger(__name__)

page: Pages = Pages.FSA

fsa_list: m.FunctionSetAssignmentsList = m.FunctionSetAssignmentsList()

current_fsa: m.FunctionSetAssignments = m.FunctionSetAssignments()

program_list: m.DERProgramList() = get_program_list()


ui.refreshable
def render_select():
    
    fsa_selection = [p.description for p in fsa_list.FunctionSetAssignments]
    fsa_selection.insert(0, "NEW")
    
    with ui.row():
        with ui.column():
            if not current_fsa.description:
                ui.select(fsa_selection, label="Function Set Assignments", value=fsa_selection[0]).classes("w-64")
            else:
                ui.select(fsa_selection, label="Function Set Assignments", value=current_fsa.description).classes("w-64")
                
            if current_fsa.href:
                ui.label(f"Href: {current_fsa.href}")



@ui.refreshable
def render_form():
    with ui.row():
        ui.input("description").classes("w-64") \
            .bind_value(current_fsa, "description")
    ui.separator()
    
    with ui.row():
        with ui.column():
            pass
                
    with ui.row():
        with ui.column():
            pass

            
            
def revert_changes():
    pass

def validate_and_submit():
    """Validate the objects and submit them to the server.
    
    Returns:
        bool: True if successful, False otherwise.
    """
    global der_base, der_default
    
    if append_new := current_fsa.href is None:
        ...
        
    fsa = send_fsa(current_fsa)
    
    assert isinstance(fsa, m.FunctionSetAssignments)
    
    if append_new:
        fsa_list.FunctionSetAssignments.append(fsa)
    
    render_select.refresh()
    
    ui.notify(f"{page.value.title} saved successfully")
    
        
@ui.page(page.value.uri)
def show_fsa():
    global fsa_list
    
    response = get_fsa_list()
    if response:
        fsa_list = response
        
        
    show_global_header(Pages.FSA)
    render_select()
    render_form()
    
    ui.separator()
    with ui.row():
        ui.button("Store", on_click=validate_and_submit)
        ui.button("Revert", on_click=revert_changes)


import subprocess


def execute_command(cmds, env=None, cwd=None, logger=None, err_prefix=None) -> str:
    """ Executes a command as a subprocess
    If the return code of the call is 0 then return stdout otherwise
    raise a RuntimeError.  If logger is specified then write the exception
    to the logger otherwise this call will remain silent.
    :param cmds:list of commands to pass to subprocess.run
    :param env: environment to run the command with
    :param cwd: working directory for the command
    :param logger: a logger to use if errors occure
    :param err_prefix: an error prefix to allow better tracing through the error message
    :return: stdout string if successful
    :raises RuntimeError: if the return code is not 0 from suprocess.run
    """

    results = subprocess.run(cmds,
                             env=env,
                             cwd=cwd,
                             stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
    if results.returncode != 0:
        err_prefix = err_prefix if err_prefix is not None else "Error executing command"
        err_message = "\n{}: Below Command failed with non zero exit code.\n" \
                      "Command:{} \nStderr:\n{}\n".format(err_prefix,
                                                          results.args,
                                                          results.stderr)
        if logger:
            logger.exception(err_message)
            raise RuntimeError()
        else:
            raise RuntimeError(err_message)

    return results.stdout.decode('utf-8')



from __future__ import annotations

import inspect
import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, TypeVar, Union

import yaml
from dataclasses_json import dataclass_json

__all__ = ["ServerConfiguration", "ReturnValue"]

import json

try:
    from gridappsd.field_interface import MessageBusDefinition
except ImportError as ex:
    pass

import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.server.exceptions import NotFoundError
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)

D = TypeVar("D")


@dataclass
class ReturnValue:
    success: bool
    an_object: D
    was_update: bool
    location: str = None

    def get(self, datatype: D) -> D:
        return self.an_object


class InvalidConfigFile(Exception):
    pass


@dataclass
class FSAConfiguration:
    description: str
    programs: List[ProgramConfiguration] = field(default_factory=list)


@dataclass
class DERConfiguration:
    #capabilities:
    modesSupported: str
    type: int


@dataclass
class DeviceConfiguration:
    id: str = None
    lfdi: Lfdi | None = None
    post_rate: int = 900
    pin: int = None
    poll_rate: int = 900
    fsas: List[str] = field(default_factory=list)
    ders: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.id.__hash__()


@dataclass
class CurveConfiguration:

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ControlBaseConfiguration:

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ControlConfiguration:
    description: str = None
    base: Dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, env):
        return cls(
            **{
                k: v
                for k, v in env.items() if k in inspect.signature(m.DERControl).parameters
                or k in inspect.signature(cls).parameters
            })

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class ProgramConfiguration:

    default_control: str = None
    controls: List[str] = field(default_factory=list)
    curves: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass_json
@dataclass
class GridappsdConfiguration:
    model_name: str
    default_pin: str
    publish_interval_seconds: int
    house_named_inverters_regex: str | None = None
    utility_named_inverters_regex: str | None = None
    model_dict_file: str | None = None
    address: str = 'localhost'
    port: int = 61613
    username: str = 'system'
    password: str = 'manager'
    field_bus_def: MessageBusDefinition | str | None = None
    feeder_id_file: Optional[str] = None
    feeder_id: Optional[str] = None
    simulation_id_file: Optional[str] = None

    @property
    def full_address(self):
        return f"tcp://{self.address}:{self.port}"

    def __post_init__(self):
        if self.field_bus_def is not None:
            if isinstance(self.field_bus_def, str):
                fb = json.loads(self.field_bus_def)
            else:
                fb = self.field_bus_def

            if is_ot_bus := fb.get('is_ot_bus', True):
                fb['connection_type'] = 'CONNECTION_TYPE_GRIDAPPSD'
                fb['connection_args'] = dict(
                    GRIDAPPSD_ADDRESS=self.full_address,
                    GRIDAPPSD_USER=self.username,
                    GRIDAPPSD_PASSWORD=self.password
                )

            else:
                assert fb['connection_args']
                assert fb['connection_type']

            # TODO: Error in gridappsd-python library the spelling is definately incorrect.
            fb['conneciton_args'] = fb.pop('connection_args')
            assert fb['id']

            self.field_bus_def = MessageBusDefinition(**fb)


@dataclass
class ProgramList:
    name: str
    programs: List[m.DERProgram]


@dataclass
class ServerConfiguration:
    openssl_cnf: str

    tls_repository: str

    server: str
    port: int

    service_name: str = "IEEE_2030_5"
    simulation_id: str | None = None

    ui_port: int = None

    include_default_der_on_all_devices: bool = True
    include_default_der_program_on_ders: bool = True

    default_program: m.DERProgram | None = None
    default_der_control: m.DefaultDERControl | None = None

    cleanse_storage: bool = True
    storage_path: str = None

    log_event_list_poll_rate: int = 900
    device_capability_poll_rate: int = 900
    mirror_usage_point_post_rate: int = 300
    end_device_list_poll_rate: int = 86400    # daily check-in

    generate_admin_cert: bool = False
    lfdi_client: str | None = None

    fsas: List[FSAConfiguration] = field(default_factory=list)
    programs: List[ProgramConfiguration] = field(default_factory=list)
    devices: List[DeviceConfiguration] = field(default_factory=list)
    ders: List[DERConfiguration] = field(default_factory=list)
    curves: List[CurveConfiguration] = field(default_factory=list)

    server_mode: Union[
        Literal["enddevices_create_on_start"],
        Literal["enddevices_register_access_only"]] = "enddevices_register_access_only"

    lfdi_mode: Union[
        Literal["lfdi_mode_from_file"],
        Literal["lfdi_mode_from_cert_fingerprint"]] = "lfdi_mode_from_cert_fingerprint"

    # programs: List[DERProgramConfiguration] = field(default_factory=list)
    # controls: List[DERControlConfiguration] = field(default_factory=list)
    # curves: List[DERCurveConfiguration] = field(default_factory=list)
    # events: List[Dict] = field(default_factory=list)

    # # map into program_lists array for programs for specific
    # # named list.
    # programs_map: Dict[str, int] = field(default_factory=dict)
    # program_lists: List[ProgramList] = field(default_factory=list)
    # fsa_list: List[FunctionSetAssignments] = field(default_factory=list)
    # curve_list: List[DERCurve] = field(default_factory=list)

    proxy_hostname: Optional[str] = None
    gridappsd: Optional[GridappsdConfiguration] = None
    # DefaultDERControl: Optional[DefaultDERControl] = None
    # DERControlList: Optional[DERControl] = field(default=list)

    @property
    def server_hostname(self) -> str:
        server = self.server
        if self.port:
            server = server + f":{self.port}"

        return server

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    @classmethod
    def load(cls, file: Path) -> ServerConfiguration:
        if not file.exists():
            raise InvalidConfigFile(f"File does not exist: {file}")
        return cls.from_dict(yaml.safe_load(file.read_text()))

    def __post_init__(self):
        # self.curves = [DERCurveConfiguration.from_dict(x) for x in self.curves]
        # self.controls = [DERControlConfiguration.from_dict(x) for x in self.controls]
        # self.programs = [DERProgramConfiguration.from_dict(x) for x in self.programs]

        if self.devices is None:
            self.devices = []
        else:
            self.devices = [DeviceConfiguration.from_dict(x) for x in self.devices]

        if self.default_program:
            # Get DefaultDERControl off of the default program and bulid the base.
            if 'DefaultDERControl' in self.default_program:
                self.default_der_control = m.DefaultDERControl(
                    **{k: v for k, v in self.default_program.items() if k in inspect.signature(m.DefaultDERControl).parameters})

                if 'DERControlBase' in self.default_program['DefaultDERControl']:
                    cb = self.default_program['DefaultDERControl']['DERControlBase']
                    self.default_der_control.DERControlBase = m.DERControlBase(**{k: v for k, v in cb.items() if k in inspect.signature(m.DERControlBase).parameters})


            # Populate from the default_program dictionary the keys of the configuration file.
            self.default_program = m.DERProgram(
                **{k: v
                   for k, v in self.default_program.items() if k in inspect.signature(m.DERProgram).parameters})

        if self.gridappsd:
            self.gridappsd = GridappsdConfiguration.from_dict(self.gridappsd)

            # TODO Configuration for field bus here
            # if Path(self.gridappsd.feeder_id_file).exists():
            #     self.gridappsd.feeder_id = Path(self.gridappsd.feeder_id_file).read_text().strip()
            # if Path(self.gridappsd.simulation_id_file).exists():
            #     self.gridappsd.simulation_id = Path(
            #         self.gridappsd.simulation_id_file).read_text().strip()

            # if not self.gridappsd.feeder_id:
            #     raise ValueError(
            #         "Feeder id from gridappsd not found in feeder_id_file nor was specified "
            #         "in gridappsd config section.")

            # # TODO: This might not be the best place for this manipulation
            # self.gridappsd.field_bus_def = MessageBusDefinition.load(
            #     self.gridappsd.field_bus_def)
            # self.gridappsd.field_bus_def.id = self.gridappsd.feeder_id

            # _log.info("Gridappsd Configuration For Simulation")
            # _log.info(f"feeder id: {self.gridappsd.feeder_id}")
            # if self.gridappsd.simulation_id:
            #     _log.info(f"simulation id: {self.gridappsd.simulation_id}")
            # else:
            #     _log.info("no simulation id")
            # _log.info("x" * 80)

        # if self.field_bus_def:
        #     self.field_bus_def = MessageBusDefinition.load(self.field_bus_def)

    def get_device_pin(self, lfdi: Lfdi, tls_repo: TLSRepository) -> int:
        for d in self.devices:
            test_lfdi = tls_repo.lfdi(d.id)
            if test_lfdi == int(lfdi):
                return d.pin
        raise NotFoundError(f"The device_id: {lfdi} was not found.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ieee_2030_5',
 'ieee_2030_5.adapters',
 'ieee_2030_5.client',
 'ieee_2030_5.data',
 'ieee_2030_5.models',
 'ieee_2030_5.persistance',
 'ieee_2030_5.server',
 'ieee_2030_5.simulation',
 'ieee_2030_5.types_',
 'ieee_2030_5.utils',
 'ieee_2030_5_gui',
 'ieee_2030_5_gui.pages',
 'ieee_2030_5_gui.views']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.3,<3.0.0',
 'blinker>=1.5,<2.0',
 'click>=8.1.3,<9.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'docutils!=0.21',
 'flask-session>=0.5.0,<0.6.0',
 'flask-talisman>=1.0.0,<2.0.0',
 'flet>=0.22.1,<0.23.0',
 'gevent>=21.12.0,<22.0.0',
 'grequests>=0.6.0,<0.7.0',
 'gridappsd-field-bus>=2024.4.1a0,<2025.0.0',
 'gridappsd-python>=2024.4.1a0,<2025.0.0',
 'pickleDB>=0.9.2,<0.10.0',
 'pyOpenSSL>=22.0.0,<23.0.0',
 'pyyaml>=6.0.1,<7.0.0',
 'simplekv>=0.14.1,<0.15.0',
 'trio>=0.21.0,<0.22.0',
 'tzlocal>=4.2,<5.0',
 'xsdata[lxml]>=22.5,<23.0']

entry_points = \
{'console_scripts': ['2030_5_cert = ieee_2030_5.certs:_main',
                     '2030_5_ctl = ieee_2030_5.control:_main',
                     '2030_5_gridappsd = ieee_2030_5.config_setup:_main',
                     '2030_5_gui = ieee_2030_5_gui.__main__:_main',
                     '2030_5_proxy = ieee_2030_5.basic_proxy:_main',
                     '2030_5_server = ieee_2030_5.__main__:_main',
                     '2030_5_shutdown = ieee_2030_5.__main__:_shutdown']}

setup_kwargs = {
    'name': 'gridappsd-2030-5',
    'version': '0.0.2a28',
    'description': '',
    'long_description': '# GridAPPS-D IEEE 2030.5 Server\n\n## Overview\n\nThe GridAPPS-D IEEE 2030.5 Server implements the Common Smart Inverter Profile (CSIP).  The server\ncan work in both in-band and out-of-band registration models detailed in  CCIP Implementation Guide v2\nsection 6.1.3 and 6.1.4 respectively.  \n\n## Setup\n\n## Installing from pypi\n\nThe recommended way to install this project from pypi is in a virtual environment.  Create a virtual environment and install\n2030.5 server as follows.\n\n```shell\n# creates an environment \'env\' in the current directory\npython3 -m venv env\n\n# Activate the environment in the current shell\nsource env/bin/activate\n\n# Install 2030.5 server\npip install gridappsd_2030_5\n```\n\n## Configuration\n\nThe server requires two configuration files.  The first is for generation of certificates and keys for the system (openssl.cnf).  Currently\nwe shell out to openssl for key/cert/ca generation during runtime.  Each time the server is started up it will attempt to regenerate\nkey/certs/ca unless --no-create-certs is passed to the server startup method.  The second is the configuration for the server itself.  \nThe configuration file holds the definitions for controls, der, end devices and other settings that will be used during the runtime\nof the server.\n\nThe following files, openssl.cnf and config.yml, should be placed in the same directory where the server is started from.\n\n### Example openssl.cnf\n\n```ini\n# OpenSSL example configuration file.\n# This is mostly being used for generation of certificate requests.\n#\n\n# This definition stops the following lines choking if HOME isn\'t\n# defined.\nHOME                    = .\nRANDFILE                = $ENV::HOME/.rnd\n\n# Extra OBJECT IDENTIFIER info:\n#oid_file               = $ENV::HOME/.oid\noid_section             = new_oids\n\n# To use this configuration file with the "-extfile" option of the\n# "openssl x509" utility, name here the section containing the\n# X.509v3 extensions to use:\n# extensions            =\n# (Alternatively, use a configuration file that has only\n# X.509v3 extensions in its main [= default] section.)\n\n[ new_oids ]\n\n# We can add new OIDs in here for use by \'ca\', \'req\' and \'ts\'.\n# Add a simple OID like this:\n# testoid1=1.2.3.4\n# Or use config file substitution like this:\n# testoid2=${testoid1}.5.6\n\n# Policies used by the TSA examples.\ntsa_policy1 = 1.2.3.4.1\ntsa_policy2 = 1.2.3.4.5.6\ntsa_policy3 = 1.2.3.4.5.7\n\n####################################################################\n[ ca ]\ndefault_ca      = CA_default            # The default ca section\n\n[ CA_default ]\ndir             = $ENV::HOME/tls             # Where everything is kept\ncerts           = $dir/certs            # Where the issued certs are kept\ndatabase        = $dir/index.txt        # database index file.\n                                        # several certs with same subject.\nnew_certs_dir   = $dir/certs            # default place for new certs.\ncertificate     = $dir/certs/ec-cacert.pem       # The CA certificate\nserial          = $dir/serial           # The current serial number\ncrlnumber       = $dir/crlnumber        # the current crl number\n                                        # must be commented out to leave a V1 CRL\nprivate_key     = $dir/private/ec-cakey.pem # The private key\n\nname_opt        = ca_default            # Subject Name options\ncert_opt        = ca_default            # Certificate field options\n\ndefault_days    = 365                   # how long to certify for\ndefault_crl_days= 30                    # how long before next CRL\ndefault_md      = sha256                # use SHA-256 by default\npreserve        = no                    # keep passed DN ordering\npolicy          = policy_match\n\n# For the CA policy\n[ policy_match ]\ncountryName             = optional\nstateOrProvinceName     = optional\norganizationName        = optional\norganizationalUnitName  = optional\ncommonName              = supplied\nemailAddress            = optional\n\n[ policy_anything ]\ncountryName             = optional\nstateOrProvinceName     = optional\nlocalityName            = optional\norganizationName        = optional\norganizationalUnitName  = optional\ncommonName              = supplied\nemailAddress            = optional\n\n####################################################################\n[ req ]\ndefault_bits            = 2048\ndefault_md              = sha256\ndefault_keyfile         = privkey.pem\ndistinguished_name      = req_distinguished_name\nattributes              = req_attributes\nx509_extensions = v3_ca # The extentions to add to the self signed cert\n\n[ req_distinguished_name ]\ncountryName                     = Country Name (2 letter code)\ncountryName_default             = IN\ncountryName_min                 = 2\ncountryName_max                 = 2\nstateOrProvinceName             = State or Province Name (full name)\nstateOrProvinceName_default     = Some-State\nlocalityName                    = Locality Name (eg, city)\nlocalityName_default            = BANGALORE\n0.organizationName              = Organization Name (eg, company)\n0.organizationName_default      = GoLinuxCloud\norganizationalUnitName          = Organizational Unit Name (eg, section)\ncommonName                      = Common Name (eg, your name or your server\\\'s hostname)\ncommonName_max                  = 64\nemailAddress                    = Email Address\nemailAddress_max                = 64\n\n[ req_attributes ]\nchallengePassword               = A challenge password\nchallengePassword_min           = 4\nchallengePassword_max           = 20\nunstructuredName                = An optional company name\n\n\n[ v3_req ]\n# Extensions to add to a certificate request\nbasicConstraints = CA:FALSE\nkeyUsage = nonRepudiation, digitalSignature, keyEncipherment\n\n[ v3_ca ]\n# Extensions for a typical CA\nsubjectKeyIdentifier=hash\nauthorityKeyIdentifier=keyid:always,issuer\nbasicConstraints = critical,CA:true\n\n[ crl_ext ]\n# issuerAltName=issuer:copy\nauthorityKeyIdentifier=keyid:always\n```\n\n```yaml\n\n### Example config.yml\n---\n#server_hostname: 0.0.0.0:8443\n\nserver: 127.0.0.1\n# Only include if we need to have dcap be available\n# http_port: 8080\nhttps_port: 7443\n\nproxy_hostname: 0.0.0.0:8443\n#server_hostname: 0.0.0.0:7443\n# server_hostname: gridappsd_dev_2004:7443\n\ntls_repository: "./tls"\nopenssl_cnf: "openssl.cnf"\n\n#server_mode: enddevices_register_access_only\nserver_mode: enddevices_create_on_start\n\n# lfdi_mode: Determines what piece of information is used to calculate the lfdi\n#\n# Options:\n#   lfdi_mode_from_file             - sha256 hash of certificate file\'s content.\n#   lfdi_mode_from_cert_fingerprint - sha256 hash of the certificates fingerprint.\n#\n# default: lfdi_mode_from_cert_fingerprint\n#lfdi_mode: lfdi_mode_from_file\nlfdi_mode: lfdi_mode_from_cert_fingerprint\n\n# Create an administrator certificate that can be used from\n# browser/api to connect to the platform.\ngenerate_admin_cert: True\n\nlog_event_list_poll_rate: 60\ndevice_capability_poll_rate: 60\n\n# End Device\ndevices:\n  # SolarEdge SE6000H HD-Wave SetApp Enabled Inverter\n  - id: dev1\n    # DeviceCategoryType from ieee_2030_5.models.device_category\n    deviceCategory: FUEL_CELL\n    pin: 111115\n\n    programs:\n      - description: Program 1\n\n    # nameplate:\n    #   rtgMaxW: 6000\n    ders:\n      - capabilities:\n        modesSupported: "1110000000000000"\n        type: 83\n\n      - capabilities:\n        # Bitmask with the following structure.\n        # Indication of support for each control mode function DERCapability::modesSupported\n        #\n        # 0 - Charge mode\n        # 1 - Discharge mode\n        # 2 - opModConnect (Connect / Disconnect -\n        # implies galvanic isolation)\n        # 3 - opModEnergize (Energize / De-Energize)\n        # 4 - opModFixedPFAbsorbW (Fixed Power\n        # Factor Setpoint when absorbing active\n        # power)\n        # 5 - opModFixedPFInjectW (Fixed Power\n        # Factor Setpoint when injecting active\n        # power)\n        # 6 - opModFixedVar (Reactive Power\n        # Setpoint)\n        # 7 - opModFixedW (Charge / Discharge\n        # Setpoint)\n        # 8 - opModFreqDroop (Frequency-Watt\n        # Parameterized Mode)\n        # 9 - opModFreqWatt (Frequency-Watt\n        # Curve Mode)\n        # 10 - opModHFRTMayTrip (High Frequency\n        # Ride Through, May Trip Mode)\n        # 11 - opModHFRTMustTrip (High\n        # Frequency Ride Through, Must Trip Mode)\n        # 12 - opModHVRTMayTrip (High Voltage\n        # Ride Through, May Trip Mode)\n        # 13 - opModHVRTMomentaryCessation\n        # (High Voltage Ride Through, Momentary\n        # Cessation Mode)\n        # 14 - opModHVRTMustTrip (High Voltage\n        # Ride Through, Must Trip Mode)\n        # 15 - opModLFRTMayTrip (Low Frequency\n        # Ride Throu\n        modesSupported: "1110000000000000"\n\n        # Item type for the DER.\n        # 0 - Not applicable / Unknown\n        # 1 - Virtual or mixed DER\n        # 2 - Reciprocating engine\n        # 3 - Fuel cell\n        # 4 - Photovoltaic system\n        # 5 - Combined heat and power\n        # 6 - Other generation system\n        # 80 - Other storage system\n        # 81 - Electric vehicle\n        # 82 - EVSE\n        # 83 - Combined PV and storage\n        type: 83\n\n        # Default available nameplate options where at a manufacturer\'s set.\n        # Active power rating in watts an unity power factor\n        rtgMaxW: 600\n\n        # Active power rating in watts at specified over-excited power factor\n        # rtgOverExcitedW:\n\n        # Over-excited power factor DERCapability::rtgOverExcitedPF\n        # rtgOverExcitedPF:\n\n        # Active power rating in watts at specified under-excited power factor DERCapability::rtgUnderExcitedW\n        # rtgUnderExcitedW:\n\n        # Under-excited power factor DERCapability::rtgUnderExcitedPF\n        # rtgUnderExcitedPF:\n\n        # Maximum apparent power rating in voltamperes DERCapability::rtgMaxVA\n        rtgMaxVA: 600\n\n        # Indication of reactive power and voltage/power control capability DERCapability::rtgNormalCategory\n        rtgNormalCategory: 1\n\n        # Indication of voltage and frequencyride-through capability category I, II, or III DERCapability::rtgAbnormalCategory\n        rtgAbnormalCategory: 1\n\n        # Maximum injected reactive power rating in vars DERCapability::rtgMaxVar\n        rtgMaxVar: 600\n\n        # Maximum absorbed reactive power rating in vars DERCapability::rtgMaxVarNeg\n        rtgMaxVarNeg: 600\n\n        # Maximum active power charge rating in watts DERCapability::rtgMaxChargeRateW\n        rtgMaxChargeRateW: 600\n\n        # Maximum apparent power charge rating in voltamperes; may differ from the apparent power maximum rating\n        # DERCapability::rtgMaxChargeRateVA\n        rtgMaxChargeRateVA: 600\n\n        # Nominal ac voltage rating in rms volts DERCapability::rtgVNom\n        rtgVNom: 120\n\n        # Maximum ac voltage rating in rms volts DERCapability::rtgMaxV\n        rtgMaxV: 128\n\n        # Minimum ac voltage rating in rms volts DERCapability::rtgMinV\n        rtgMinV: 116\n\n        # Reactive susceptance that remains connected to the Area EPS in the cease to energize and trip state\n        # DERCapability::rtgReactiveSusceptance\n        # rtgReactiveSusceptance:\n\n        # # Manufacturer DeviceInformation::mfID\n        # mfID:\n\n        # # Model DeviceInformation::mfModel\n        # mfModel:\n\n        # # Serial number DeviceInformation::mfSerNum\n        # mfSerNum:\n\n        # # Version DeviceInformation::mfHwVer DeviceInformation::swVer\n        # mfHwVer:\n        # swVer:\n\n  - id: dev2\n    deviceCategory: FUEL_CELL\n    pin: 12345\n    nameplate:\n\nprograms:\n  - description: Program 1\n    default_control: Control 1\n    controls:\n      - Control 2\n      - Control 3\n    curves:\n      - Curve 1\n    primacy: 89\n\ncontrols:\n  - description: Control 1\n    setESDelay: 30\n    base:\n      opModConnect: True\n      opModMaxLimW: 9500\n\n      # setESHighFreq: UInt16 [0..1]\n      # setESHighVolt: Int16 [0..1]\n      # setESLowFreq: UInt16 [0..1]\n      # setESLowVolt: Int16 [0..1]\n      # setESRampTms: UInt32 [0..1]\n      # setESRandomDelay: UInt32 [0..1]\n      # setGradW: UInt16 [0..1]\n      # setSoftGradW: UInt16 [0..1]\n  - description: Control 2\n  - description: Control 3\n\nevents:\n  - control: 0\n\ncurves:\n  # Each element will can have the following structure.\n  # autonomousVRefEnable: If the curveType is opModVoltVar, then\n  #   this field MAY be present. If the curveType is not opModVoltVar,\n  #   then this field SHALL NOT be present. Enable/disable autonomous\n  #   vRef adjustment. When enabled, the Volt-Var curve characteristic\n  #   SHALL be adjusted autonomously as vRef changes and\n  #   autonomousVRefTimeConstant SHALL be present. If a DER is able to\n  #   support Volt-Var mode but is unable to support autonomous vRef\n  #   adjustment, then the DER SHALL execute the curve without\n  #   autonomous vRef adjustment. If not specified, then the value is\n  #   false.\n  # autonomousVRefTimeConstant: If the curveType is opModVoltVar,\n  #   then this field MAY be present. If the curveType is not\n  #   opModVoltVar, then this field SHALL NOT be present. Adjustment\n  #   range for vRef time constant, in hundredths of a second.\n  # creationTime: The time at which the object was created.\n  # CurveData:\n  # curveType: Specifies the associated curve-based control mode.\n  # openLoopTms: Open loop response time, the time to ramp up to\n  #   90% of the new target in response to the change in voltage, in\n  #   hundredths of a second. Resolution is 1/100 sec. A value of 0 is\n  #   used to mean no limit. When not present, the device SHOULD\n  #   follow its default behavior.\n  # rampDecTms: Decreasing ramp rate, interpreted as a percentage\n  #   change in output capability limit per second (e.g. %setMaxW /\n  #   sec).  Resolution is in hundredths of a percent/second. A value\n  #   of 0 means there is no limit. If absent, ramp rate defaults to\n  #   setGradW.\n  # rampIncTms: Increasing ramp rate, interpreted as a percentage\n  #   change in output capability limit per second (e.g. %setMaxW /\n  #   sec).  Resolution is in hundredths of a percent/second. A value\n  #   of 0 means there is no limit. If absent, ramp rate defaults to\n  #   rampDecTms.\n  # rampPT1Tms: The configuration parameter for a low-pass filter,\n  #   PT1 is a time, in hundredths of a second, in which the filter\n  #   will settle to 95% of a step change in the input value.\n  #   Resolution is 1/100 sec.\n  # vRef: If the curveType is opModVoltVar, then this field MAY be\n  #   present. If the curveType is not opModVoltVar, then this field\n  #   SHALL NOT be present. The nominal AC voltage (RMS) adjustment to\n  #   the voltage curve points for Volt-Var curves.\n  # xMultiplier: Exponent for X-axis value.\n  # yMultiplier: Exponent for Y-axis value.\n  # yRefType: The Y-axis units context.\n  # Each curve MUST have between 1 and 10 elements in the curve_data list.\n  #\n  # DERCurve Type for each curve.\n  # 0 - opModFreqWatt (Frequency-Watt Curve Mode)\n  # 1 - opModHFRTMayTrip (High Frequency Ride Through, May Trip Mode)\n  # 2 - opModHFRTMustTrip (High Frequency Ride Through, Must Trip Mode)\n  # 3 - opModHVRTMayTrip (High Voltage Ride Through, May Trip Mode)\n  # 4 - opModHVRTMomentaryCessation (High Voltage Ride Through, Momentary Cessation\n  # Mode)\n  # 5 - opModHVRTMustTrip (High Voltage Ride Through, Must Trip Mode)\n  # 6 - opModLFRTMayTrip (Low Frequency Ride Through, May Trip Mode)\n  # 7 - opModLFRTMustTrip (Low Frequency Ride Through, Must Trip Mode)\n  # 8 - opModLVRTMayTrip (Low Voltage Ride Through, May Trip Mode)\n  # 9 - opModLVRTMomentaryCessation (Low Voltage Ride Through, Momentary Cessation\n  # Mode)\n  # 10 - opModLVRTMustTrip (Low Voltage Ride Through, Must Trip Mode)\n  # 11 - opModVoltVar (Volt-Var Mode)\n  # 12 - opModVoltWatt (Volt-Watt Mode)\n  # 13 - opModWattPF (Watt-PowerFactor Mode)\n  # 14 - opModWattVar (Watt-Var Mode)\n  - description: Curve 1\n    curveType: opModVoltVar\n    CurveData:\n      - xvalue: 5\n        yvalue: 5\n\n  - description: Curve 2\n    curveType: opModFreqWatt\n    CurveData:\n      # exitation is only available if yvalue is power factor\n      - exitation: 10\n        xvalue: 5\n        yvalue: 5\n```\n\n### Starting the server\n\nAfter installing one can start the server by the folloing command:\n\n```bash\n2030_5_server --no-validate config.yml\n```\n\nTo learn about other options please use the help module.\n\n```bash\n2030_5_server --help\n```\n\n## Installing from source\n\nThe installation requires poetry version 1.2 or greater.  <https://python-poetry.org/docs/#installation>\n\n 1. Clone the repository\n 2. run `poetry install` from the root of the repository directory.\n\n### Clone the repository\n\n```shell\ngit clone https://github.com/GRIDAPPSD/gridappsd-2030_5 -b develop\ncd gridappsd-2030_5\n```\n\n### Install Requirements\n\n```shell\npoetry install\n```\n\n### Run the Server\n\n```shell\nusage: 2030_5_server [-h] [--no-validate] [--no-create-certs] [--debug] config\n\npositional arguments:\n  config             Configuration file for the server.\n\noptional arguments:\n  -h, --help         show this help message and exit\n  --no-validate      Allows faster startup since the resolving of addresses for devices is not done.\n  --no-create-certs  If specified certificates for for client and server will not be created.\n  --debug            Put server in debug mode (more logging)\n```\n\n### Using the 2030.5 proxy\n\nThe proxy is used to keep a http 1.1 connection alive rather than doing the tls setup\nmore than one time.\n\n```shell\nusage: 2030_5_proxy [-h] [--debug] config\n\npositional arguments:\n  config      Configuration file for the server.\n\noptional arguments:\n  -h, --help  show this help message and exit\n  --debug     Turns debugging on for logging of the proxy.\n```\n\n## Client Connectivity\n\nThe server will expose an endpoint of beginning with <https://myserver/dcap>.  From there\na client will be able to traverse and do any PUT, POST, GET, and DELETE operations specified\nin the 2030.5 test procedures.\n\n## Function Sets Implemented\n',
    'author': 'C. Allwardt',
    'author_email': '3979063+craig8@users.noreply.github.com',
    'maintainer': 'C. Allwardt',
    'maintainer_email': '3979063+craig8@users.noreply.github.com',
    'url': 'https://github.com/GRIDAPPSD/gridappsd-2030_5',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Optional
from ieee_2030_5.models.sep import (
    DER,
    IdentifiedObject,
    Link,
    Resource,
    SubscribableList,
)

__NAMESPACE__ = "epri:derforecast:ns"


@dataclass
class ForecastNumericType:
    """
    Real number expressed as an integer and power-of-ten-multiplier.

    :ivar value: Value expressed as integer
    :ivar multiplier: Multiplier for value. Multiply value by 10^this.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    value: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    multiplier: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )


@dataclass
class DERForecastLink(Link):
    """
    SHALL contain a Link to an instance of DERForecast.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    postRate: int = field(
        default=900,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class ForecastParameter:
    """
    Object holding forecast for a single parameter.

    :ivar name: Name of the paramater
    :ivar forecast: Forecast for the parameter named.
    :ivar sigma: Standard deviation for the parameter named.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    name: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    forecast: Optional[ForecastNumericType] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    sigma: Optional[ForecastNumericType] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class DERFlexibility(DER):
    """
    Extends sep DER to include DERForecastLink.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    DERForecastLink: Optional[DERForecastLink] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class ForecastParameterSet(Resource):
    """
    A set of forecasts.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    ForecastParameter: List[ForecastParameter] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class ForecastParameterSetList(SubscribableList):
    """
    A List element to hold ForecastParameterSet  objects.
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    ForecastParameterSet: List[ForecastParameterSet] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class DERForecast(IdentifiedObject):
    """
    DER forecast information.

    :ivar startTime: The start time in epoch for this forecast.
    :ivar interval: Forecast interval for the included
        ForecastParameterSetList, in seconds.
    :ivar ForecastParameterSetList:
    """
    class Meta:
        namespace = "epri:derforecast:ns"

    startTime: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    interval: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "required": True,
        }
    )
    ForecastParameterSetList: Optional[ForecastParameterSetList] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )



import argparse
import hashlib
import logging
import os
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple
import shutil
import yaml
from cryptography import x509
from cryptography.hazmat.backends import default_backend

__all__ = ['TLSRepository']

from ieee_2030_5.types_ import Lfdi, PathStr
from ieee_2030_5.utils.tls_wrapper import OpensslWrapper, TLSWrap
from ieee_2030_5.utils.cryptography_wrapper import CryptographyWrapper

_log = logging.getLogger(__name__)

PRIVATE_EXTENTION = 'pem'
CERTIFICATE_EXTENSION = 'crt'

PRIVATE_EXTENTION = os.environ.get('2030_5_PRIVATE_EXTENSION', PRIVATE_EXTENTION)
CERTIFICATE_EXTENSION = os.environ.get('2030_5_PUBLIC_EXTENSION', CERTIFICATE_EXTENSION)

GLOB_PRIVATE = f'*.{PRIVATE_EXTENTION}'
GLOB_CERT = f'*.{CERTIFICATE_EXTENSION}'

def lfdi_from_fingerprint(fingerprint: str) -> Lfdi:
    fp = fingerprint.replace(":", "")
    return Lfdi(fp[:40])


def sfdi_from_lfdi(lfdi: Lfdi) -> int:
    assert len(lfdi) == 40, "lfdi must be 160-bits (40 hex characters) long."
    hex_str = str(int(lfdi[:9], 16))
    check_bit = 0
    full_sum = sum([int(x) for x in hex_str])
    while not (full_sum + check_bit) % 10 == 0:
        check_bit += 1
    return int(hex_str + str(check_bit))


class TLSRepository:

    def __init__(self,
                 repo_dir: PathStr,
                 openssl_cnffile_template: PathStr,
                 serverhost: str,
                 proxyhost: str = None,
                 clear=False,
                 **kwargs):
        if isinstance(repo_dir, str):
            repo_dir = Path(repo_dir).expanduser().resolve()
        if isinstance(openssl_cnffile_template, str):
            openssl_cnffile_template = Path(openssl_cnffile_template).expanduser().resolve()
            if not openssl_cnffile_template.exists():
                raise ValueError(f"openssl_cnffile does not exist {openssl_cnffile_template}")
        self._repo_dir = repo_dir
        self._certs_dir = repo_dir.joinpath("certs")
        self._private_dir = repo_dir.joinpath("private")
        self._combined_dir = repo_dir.joinpath("combined")
        self._openssl_cnf_file = self._repo_dir.joinpath(openssl_cnffile_template.name)
        self._common_names = {serverhost: serverhost}
        if proxyhost:
            self._common_names[proxyhost] = proxyhost
        self._client_common_name_set = set()
        
        if clear and self._repo_dir.exists():
            shutil.rmtree(self._repo_dir)

        if not self._repo_dir.exists() or not self._certs_dir.exists() or \
                not self._private_dir.exists() or not self._combined_dir.exists():
            self._certs_dir.mkdir(parents=True)
            self._private_dir.mkdir(parents=True)
            self._combined_dir.mkdir(parents=True)

        index_txt = self._repo_dir.joinpath("index.txt")
        serial = self._repo_dir.joinpath("serial")
        
        if not index_txt.exists():
            index_txt.write_text("")
        if not serial.exists():
            serial.write_text("01")

        self._current_pk: Dict[str, Path] = {}
        self._current_certs: Dict[str, Path] = {}
        # lfdi -> sfdi and sfdi -> lfdi for devices.
        self._devices: Dict[str, str] = {}
        
        new_contents = openssl_cnffile_template.read_text().replace(
            "dir = REPLACE_WITH_REPO_PATH", f"dir = {repo_dir}")
        self._openssl_cnf_file.write_text(new_contents)
        self._ca_key = self._private_dir / f"ca.{PRIVATE_EXTENTION}"
        self._ca_cert = self._certs_dir / f"ca.{CERTIFICATE_EXTENSION}"
        self._serverhost = serverhost
        self._proxyhost = proxyhost

        self._tls: TLSWrap = OpensslWrapper(self._openssl_cnf_file)
        # self._cert_paths: List[Path] = []
        # self._certificate_specs: Dict[str, Dict[str, str]] = {}
        if not clear:
            
            # creating certs has something screwy so we are going
            # to create the cert_paths based upon the private key
            # files.
            for f in self._private_dir.glob(GLOB_PRIVATE):
                f = Path(f)
                self._current_pk[f.stem] = f
            
            for f in self._certs_dir.glob(GLOB_CERT):
                f = Path(f)
                self._current_certs[f.stem] = f
                        
        if not self._ca_key.exists() or not self._ca_cert.exists():
            self._tls.tls_create_private_key(self.ca_key_file)
            self._tls.tls_create_ca_certificate("ca", self.ca_key_file, self.ca_cert_file)
            self._current_pk["ca"] = self.ca_key_file
            self._current_certs["ca"] = self.ca_cert_file
            
        if not self.server_cert_file.exists() or not self.server_key_file.exists():
            self._tls.tls_create_private_key(self.server_key_file)
            self._tls.tls_create_signed_certificate(serverhost, self.ca_key_file, self.ca_cert_file,
                                                    self.server_key_file, self.server_cert_file,
                                                    as_server=True)
            self._current_pk[serverhost] = self.server_key_file
            self._current_certs[serverhost] = self.server_cert_file
        
        if proxyhost is not None and (not self.proxy_cert_file.exists() or \
            not self.proxy_key_file.exists()):
            self._tls.tls_create_private_key(self.proxy_key_file)
            self._tls.tls_create_signed_certificate(proxyhost, self.ca_key_file, self.ca_cert_file,
                                                    self.proxy_key_file, self.proxy_cert_file,
                                                    as_server=True)
            self._current_pk[proxyhost] = self.proxy_key_file
            self._current_certs[proxyhost] = self.proxy_cert_file

        generate_admin_cert = kwargs.pop('generate_admin_cert', False)

        if generate_admin_cert:
            admin_key = self._private_dir / f"admin.{PRIVATE_EXTENTION}"
            admin_cert = self._certs_dir / f"admin.{CERTIFICATE_EXTENSION}"
            if not admin_key.exists():
                self._tls.tls_create_private_key(admin_key)
                self.create_cert(admin_cert.stem)
                self._current_pk["admin"] = admin_key
                self._current_certs["admin"] = admin_cert
                
        

        for crt in self._current_certs:
            if crt not in (serverhost, proxyhost, "ca", "admin"):
                self._devices[crt] = (self.lfdi(crt), self.sfdi(crt))

        assert len(self._current_pk) == len(self._current_certs)
        
        if len(kwargs) > 0:
            raise ValueError(f"Not all kwargs used: {kwargs.keys()}")

    def __create_ca__(self):
        self._tls.tls_create_private_key(self._ca_key)
        self._tls.tls_create_ca_certificate("ca", self._ca_key, self._ca_cert)
        self._tls.tls_create_pkcs23_pem_and_cert(self._ca_key, self._ca_cert,
                                                 self.__get_combined_file__("ca"))
        self._current_certs["ca"] = self.ca_cert_file
        self._current_pk["ca"] = self.ca_key_file

    def has_device(self, common_name: str) -> bool:
        return common_name in self._devices
        
    def create_cert(self, common_name: str, as_server: bool = False):

        if not self.__get_key_file__(common_name).exists():
            self._tls.tls_create_private_key(self.__get_key_file__(common_name))
            self._current_pk[common_name] = self.__get_key_file__(common_name)

        self._tls.tls_create_signed_certificate(common_name, self._ca_key, self._ca_cert,
                                                self.__get_key_file__(common_name),
                                                self.__get_cert_file__(common_name), as_server)
        self._current_certs[common_name] = self.__get_cert_file__(common_name)
        
        self._tls.tls_create_pkcs23_pem_and_cert(self.__get_key_file__(common_name),
                                                 self.__get_cert_file__(common_name),
                                                 self.__get_combined_file__(common_name))

        # self._common_names[common_name] = common_name
        # self._cert_paths.append(self.__get_cert_file__(common_name=common_name))
        # self._certificate_specs[common_name] = dict(common_name=common_name,
        #                                             lFDI=self.lfdi(common_name),
        #                                             path=self.__get_cert_file__(common_name).as_posix())
        

    def lfdi(self, device_id: str) -> Lfdi:
        """
        Using the fingerprint of the certifcate return the left truncation of 160 bits with no check digit.
        Example:
          From:
            3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A
          Return:
            3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5
            as an integer.
        """
        # 160 / 4 == 40
        fp = self.fingerprint(device_id, True)
        return Lfdi(lfdi_from_fingerprint(fp))

    def sfdi(self, device_id: str) -> int:
        lfdi_ = self.lfdi(device_id)
        return sfdi_from_lfdi(lfdi_)

    def fingerprint(self, device_id: str, without_colan: bool = True) -> str:
        if os.environ.get('IEEE_2030_5_CERT_FROM_COMBINED_FILE'):
            # _log.debug("Using hash from combined file.")
            value = Path(self.__get_combined_file__(device_id)).read_text()
            value = hashlib.sha256(value.encode('utf-8')).hexdigest()
        else:
            value = self._tls.tls_get_fingerprint_from_cert(self.__get_cert_file__(device_id))
        if without_colan:
            value = value.replace(":", "")
        if "=" in value:
            value = value.split("=")[1]
        assert isinstance(value, str)
        return value

    def get_common_name(self, device_id: str) -> x509:
        pem_data = Path(self.__get_cert_file__(device_id)).read_bytes()
        cert = x509.load_pem_x509_certificate(pem_data, default_backend())
        return cert.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0].value

    def get_file_pair(self, device_id: str) -> Tuple[str, str]:
        """ Get cert, key from the repository based on passed device_id"""
        return (self.__get_cert_file__(device_id).as_posix(),
                self.__get_key_file__(device_id).as_posix())

    @property
    def client_list(self) -> Dict[str, Dict[str, str]]:
        # TODO: Use precalculated specs rather than this each time.
        specs: Dict[str, Dict[str, str]] = {}
        for d in self._private_dir.glob(GLOB_PRIVATE):
            
            paths = self.get_file_pair(d.stem)
            
            specs[d.stem] = {'common_name': d.stem,
                             'path': ','.join(paths),
                             'device': False}
            
            if ':' not in d.stem or 'admin' != d.stem:
                specs[d.stem]['lFID'] = self.lfdi(d.stem)
                specs[d.stem]['device'] = True
                         
        return specs

    @property
    def ca_key_file(self) -> Path:
        return self._ca_key

    @property
    def ca_cert_file(self) -> Path:
        return self._ca_cert

    @property
    def proxy_key_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_key_file__(self._proxyhost)

    @property
    def proxy_cert_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_cert_file__(self._proxyhost)

    @property
    def server_key_file(self) -> Path:
        return self.__get_key_file__(self._serverhost)

    @property
    def server_cert_file(self) -> Path:
        return self.__get_cert_file__(self._serverhost)

    def find_device_id_from_sfdi(self, sfdi: int) -> Optional[str]:
        """
        Searches the certificate paths for a device id that maps to the sfdi passed into the method.
        Args:
            sfdi:

        Returns:

        """
        device_id = None
        _log.debug(f"Attempting to find sfid: {sfdi}")
        for d in self._certs_dir.glob(GLOB_CERT):
            try:
                if sfdi == self.sfdi(d.stem):
                    device_id = d.stem
                    break
            except FileNotFoundError:
                pass
        return device_id

    def __get_cert_file__(self, common_name: str) -> Path:
        return self._certs_dir.joinpath(f"{common_name}.{CERTIFICATE_EXTENSION}")

    def __get_key_file__(self, common_name: str) -> Path:
        return self._private_dir.joinpath(f"{common_name}.{PRIVATE_EXTENTION}")

    def __get_combined_file__(self, common_name: str) -> Path:
        return self._combined_dir.joinpath(f"{common_name}-combined.{PRIVATE_EXTENTION}")


def _main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dir", help="Directory of certificates determine lfdi and sfdi from.")
    parser.add_argument("--file", help="File to determine lfdi and sfdi from.")

    opts = parser.parse_args()

    if opts.dir and opts.file:
        sys.stderr.write("Only specify dir or file.\n")
        sys.exit(1)
    elif opts.dir is None and opts.file is None:
        sys.stderr.write("Must specify either --dir or --file.\n")
        sys.exit(1)

    target = opts.file
    if opts.dir:
        target = opts.dir

    target = Path(target)
    if not target.exists():
        sys.stderr.write("Invalid file or directory refrence.\n")
        sys.exit(1)

    if target.is_dir():
        targets = target.glob("*.crt")
    else:
        targets = [target]

    for t in targets:
        fingerprint = OpensslWrapper.tls_get_fingerprint_from_cert(t)
        lfdi = lfdi_from_fingerprint(fingerprint)
        sfdi = sfdi_from_lfdi(lfdi)
        sys.stdout.write(f"certificate: {t}\n")
        sys.stdout.write(f"-" * 60 + "\n")
        sys.stdout.write(f"fingerprint: {fingerprint}\n")
        sys.stdout.write(f"lfdi: {lfdi.decode('ascii')}\n")
        sys.stdout.write(f"sfdi: {sfdi}\n\n")


if __name__ == '__main__':

    logging.basicConfig(level=logging.DEBUG)

    # fingerprint = "3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A".replace(
    #     "-", "")
    # lfdi = lfdi_from_fingerprint(fingerprint)
    # sfdi = sfdi_from_lfdi(lfdi)
    # print(f"fingerprint: {fingerprint}")
    # print(f"lfdi: {lfdi}")
    # print(f"sfdi: {sfdi}")

    # fingerprint = "B5:65:B2:C4:D4:22:59:72:58:6E:4E:E2:B1:F2:98:D4:20:62:15:DB:53:49:AB:45:2F:D2:8F:BC:62:2C:28:1D".replace(
    #     ":", "")
    # lfdi = lfdi_from_fingerprint(fingerprint)
    # sfdi = sfdi_from_lfdi(lfdi)
    # print(f"fingerprint: {fingerprint}")
    # print(f"lfdi: {lfdi}")
    # print(f"sfdi: {sfdi}")

    #
    # tlsrepo = TLSRepository(repo_dir="~/tls",
    #                         openssl_cnffile_template="../openssl.cnf",
    #                         clear=False,
    #                         serverhost="gridappsd_dev_2004:8443")
    # fingerprint = tlsrepo.fingerprint("dev1")
    # # fingerprint = "3F4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213B".replace("-", "")
    # print(len(fingerprint))
    # print("my lfdi: ", fingerprint[:40])
    # tlsrepo.sfdi_from_lfdi(fingerprint[:40].encode("ascii"))
    # # Each char is 4 bits so 9*4 == 36
    # print("left 36 bits: ", fingerprint[:9])
    # print("to int from fingerprint", int(fingerprint[:9], 16))
    # interum = str(int(fingerprint[:9], 16))
    # print(int(interum[-2:]))
    # add_value = 1
    # while not (int(interum[-2:]) + add_value) % 10 == 0:
    #     add_value += 1
    #
    # print(add_value)
    # interum = interum + str(add_value)
    # print(f"sfdi = ", interum)
    #
    # print(f" our sfdi: {tlsrepo.sfdi('dev1')}")
    #
    # _log.debug(f"fingerprint: {tlsrepo.fingerprint('dev1', False)}")
    #
    # _log.debug(f"dev1 lfdi: {tlsrepo.lfdi('dev1')}, sfdi: {tlsrepo.sfdi('dev1')}")



import datetime
from pathlib import Path
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography import x509
from cryptography.x509.oid import NameOID, ExtendedKeyUsageOID


from ieee_2030_5.utils import CADoesNotExist, CertExistsError, PrivateKeyDeosntExist, TLSWrap

class CryptographyWrapper(TLSWrap):
    @staticmethod
    def tls_create_private_key(file_path: Path) -> bool:
        """
        Creates a private key in the path that is specified.  The path will be overwritten
        if it already exists.

        Args:
            file_path:

        Returns:

        """
        pk = ec.generate_private_key(ec.SECP224R1(), default_backend())
        result = pk.private_bytes(encoding=serialization.Encoding.PEM,
                         format=serialization.PrivateFormat.PKCS8,
                         encryption_algorithm=serialization.NoEncryption())
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.open("wb").write(result)
        return True


    @staticmethod
    def tls_create_ca_certificate(common_name: str, private_key_file: Path, ca_cert_file: Path):
        """
        Create a ca certificate from using common name private key and ca certificate file.

        Args:
            common_name:
            private_key_file:
            ca_cert_file:

        Returns:

        """
        if ca_cert_file.exists():
            raise CertExistsError(ca_cert_file)
        
        if not private_key_file.exists():
            CryptographyWrapper.tls_create_private_key(private_key_file)
        
        pk = serialization.load_pem_private_key(
            private_key_file.read_bytes(), None, default_backend())
        
        # Create CSR for the CA Cetificate
        # csr = x509.CertificateSigningRequestBuilder().subject_name(
        #     x509.Name([
        #         x509.NameAttribute(NameOID.COMMON_NAME, "CA")
        #         # Provide various details about who we are.
        #         # x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
        #         # x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
        #         # x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
        #         # x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
        #         # x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        #     ])).add_extension(
        #         x509.SubjectAlternativeName([
        #         # Describe what sites we want this certificate for.
        #         # x509.DNSName(u"mysite.com"),
        #         # x509.DNSName(u"www.mysite.com"),
        #         x509.DNSName(common_name),
        #     ]),
        #         critical=False,
        #     # Sign the CSR with our private key.
        #     ).sign(pk, hashes.SHA256())
        
        ca_subject = x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, common_name)
        ])
        
        # # Various details about who we are. For a self-signed certificate the
        # # subject and issuer are always the same.
        # subject = issuer = x509.Name([
        #             x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
        #             x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"California"),
        #             x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
        #             x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
        #             x509.NameAttribute(NameOID.COMMON_NAME, u"mysite.com"),
        #         ])

        cert = x509.CertificateBuilder().subject_name(
                    ca_subject
                ).issuer_name(
                    ca_subject
                ).public_key(
                    pk.public_key()
                ).serial_number(
                    x509.random_serial_number()
                ).not_valid_before(
                    datetime.datetime.now(datetime.timezone.utc)
                ).not_valid_after(
                    # Our certificate will be valid for 10 days
                    datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=20*365)
                ).add_extension(
                    x509.BasicConstraints(ca=True, path_length=None), critical=True
                 # Sign our certificate with our private key
                ).sign(pk, hashes.SHA256())
        
        ca_cert_file.write_bytes(cert.public_bytes(serialization.Encoding.PEM))

    @staticmethod
    def tls_create_csr(common_name: str,  private_key_file: Path, server_csr_file: Path):
        """

        Args:
            common_name:
            private_key_file:
            server_csr_file:

        Returns:

        """
        csr = x509.CertificateSigningRequest().subject_name(x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, common_name)
        ])).add_extention(
            x509.SubjectAlternativeName([
                x509.DNSName(common_name)
            ]),
            critical=True
        ).sign(key, hashes.SHA256())
        
    @staticmethod
    def tls_create_device_certificate(ipaddress: str,
                                      ca_key_file: Path,
                                      ca_cert_file: Path,
                                      private_key_file: Path,
                                      cert_file: Path):
        pass

    @staticmethod
    def tls_create_signed_certificate(common_name: str,
                                      ca_key_file: Path,
                                      ca_cert_file: Path,
                                      private_key_file: Path,
                                      cert_file: Path,
                                      as_server: bool = False):
        """

        Args:
            common_name:
            ca_key_file:
            ca_cert_file:
            private_key_file:
            cert_file:
            as_server:

        Returns:

        """
        
        if not ca_key_file.exists() or not ca_cert_file.exists():
            raise CADoesNotExist()
        
        if not private_key_file.exists():
            raise PrivateKeyDeosntExist(private_key_file)
        
        if cert_file.exists():
            raise CertExistsError(cert_file)
        
        pk = serialization.load_pem_private_key(
            private_key_file.read_bytes(), None, default_backend())
        
        signing_key = serialization.load_pem_private_key(
            ca_key_file.read_bytes(), None, default_backend())
        
        signing_cert = x509.load_pem_x509_certificate(
            ca_cert_file.read_bytes(), default_backend()
        )
        
        san = x509.SubjectAlternativeName([x509.DNSName(common_name)])
        
        builder = x509.CertificateBuilder().subject_name(
                    x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "")])
                ).issuer_name(
                    signing_cert.subject
                ).public_key(
                    pk.public_key()
                ).serial_number(
                    x509.random_serial_number()
                ).not_valid_before(
                    datetime.datetime.now(datetime.timezone.utc)
                ).not_valid_after(
                    # Our certificate will be valid for 10 days
                    datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=20*365)
                ).add_extension(
                    san, False
                 # Sign our certificate with our private key
                )
                
        # builder = builder.add_extension(
        #     x509.KeyUsage(digital_signature=True, key_encipherment=True,
        #                   content_commitment=False,
        #                   data_encipherment=False, key_agreement=False,
        #                   key_cert_sign=False,
        #                   crl_sign=False,
        #                   encipher_only=False, decipher_only=False
        #                   ),
        #         critical=True)
        
        if as_server:
            builder = builder.add_extension(
                x509.ExtendedKeyUsage((ExtendedKeyUsageOID.SERVER_AUTH,)),
                critical=False
            )
        cert = builder.sign(signing_key, hashes.SHA256())
        cert_file.write_bytes(
            cert.public_bytes(serialization.Encoding.PEM)
        )

    @staticmethod
    def tls_get_fingerprint_from_cert(cert_file: Path, algorithm: str = "sha256"):
        """

        Args:
            cert_file:
            algorithm:

        Returns:

        """
        cert = x509.load_pem_x509_certificate(cert_file.read_bytes(), default_backend())
        results = cert.fingerprint(hashes.SHA256())
        return results.hex(":")
        
    @staticmethod
    def tls_create_pkcs23_pem_and_cert(private_key_file: Path, cert_file: Path,
                                       combined_file: Path):
        """

        Args:
            private_key_file:
            cert_file:
            combined_file:

        Returns:

        """
        with combined_file.open("wb") as fp:
            fp.write(private_key_file.read_bytes() + b"\n" + 
                     cert_file.read_bytes() + b"\n")


import flet as ft

import json
from http.client import HTTPSConnection
import ssl
from typing import Callable

import logging
import os
from pathlib import Path
from fastapi.middleware.wsgi import WSGIMiddleware
from urllib.parse import urlparse

_log = logging.getLogger(__name__)

this_view: ft.Ref[ft.View] = ft.Ref()
this_page: ft.Ref[ft.Page] = ft.Ref()
admin_cert: ft.Ref[ft.TextField] = ft.Ref()
admin_key: ft.Ref[ft.TextField] = ft.Ref()
ca_cert: ft.Ref[ft.TextField] = ft.Ref()
text_response: ft.Ref[ft.Text] = ft.Ref()
col_response: ft.Ref[ft.Column] = ft.Ref()
admin_endpoint: ft.Ref[ft.Column] = ft.Ref()

def populate_tree() -> None:
    _log.debug("Populating tree")

    parsed_url = urlparse(admin_endpoint.current.value)

    _ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    _ssl_context.check_hostname = False
    _ssl_context.verify_mode = ssl.CERT_OPTIONAL    #  ssl.CERT_REQUIRED
    _ssl_context.load_verify_locations(cafile=os.path.expanduser(ca_cert.current.value))

    # Loads client information from the passed cert and key files. For
    # client side validation.
    _ssl_context.load_cert_chain(certfile=os.path.expanduser(admin_cert.current.value),
                                 keyfile=os.path.expanduser(admin_key.current.value))

    _http_conn = HTTPSConnection(host=parsed_url.hostname,
                                 port=parsed_url.port,
                                 context=_ssl_context)

    _http_conn.connect()
    _http_conn.request("GET", parsed_url.path+"/resources")
    response = _http_conn.getresponse()
    data = json.loads(response.read().decode("utf-8"))
    text_response.current.value = json.dumps(data, indent=4)
    _http_conn.close()
    this_page.current.update()

    # app.run(host=os.getenv(f"{PREFIX}_HOST"), port=os.getenv(f"{PREFIX}_PORT"), debug=True, ssl_context=(os.getenv(f"{PREFIX}_CLIENT_CERT"), os.getenv(f"{PREFIX}_CLIENT_KEY"), os.getenv(f"{PREFIX}_CA_CERT")))
    # app.run(host=os.getenv(f"{PREFIX}_HOST"), port=os.getenv(f"{PREFIX}_PORT"), debug=True, ssl_context=(os.getenv(f"{PREFIX}_CLIENT_CERT"), os.getenv(f"{PREFIX}_CLIENT_KEY"), os.getenv(f"{PREFIX}_CA_CERT"))





def show_resource_tree_view(page: ft.Page) -> ft.View:
    this_page.current = page

    this_view.current = ft.View("/", scroll=True, controls=[
        ft.AppBar(title=ft.Text("Server Control"),
                  automatically_imply_leading=False,
                  bgcolor=ft.colors.SURFACE_VARIANT),
        ft.TextField(
            label="Admin Certificate", ref=admin_cert, width=500, value="~/tls/certs/admin.crt"),
        ft.TextField(label="Admin Key", ref=admin_key, width=500, value="~/tls/private/admin.pem"),
        ft.TextField(label="CA Certificate", ref=ca_cert, width=500, value="~/tls/certs/ca.crt"),
        ft.TextField(label="Admin Endpoint",
                     ref=admin_endpoint,
                     width=500,
                     value="https://we48687:8090/admin"),
        ft.Column(ref=col_response, controls=[ft.Text(ref=text_response)], scroll=True),
        ft.ElevatedButton("Populate", on_click=lambda _: populate_tree()),
    ])

    return this_view.current



#!/usr/bin/env python3


from der_programs import program_add, programs_list
from enddevices import add_end_device, show_end_device, show_list
from nicegui import ui
from router import global_router as router
from session import backend_session, endpoint

import ieee_2030_5.models as m
from ieee_2030_5.utils import xml_to_dataclass

resp = backend_session.get(endpoint("enddevices"))
enddevices: m.EndDeviceList = xml_to_dataclass(resp.text)

@router.add('/')
async def show_one():
    ui.label('Content One').classes('text-2xl')
    

    

@router.add('/end-devices')
async def show_end_devices(index: int = None):
    # resp = requests.get("https://127.0.0.1:7443/admin/enddevices", 
    #                     cert=('/home/os2004/tls/certs/admin.pem', '/home/os2004/tls/private/admin.pem'),
    #                     verify="/home/os2004/tls/certs/ca.pem")
    resp = backend_session.get(endpoint("enddevices"))
    enddevices = xml_to_dataclass(resp.text)
    if index is None:
        for index, ed in enumerate(enddevices.EndDevice):
            router.add_route(f'/end-devices/{index}', lambda : show_end_devices(index))

        show_list(enddevices)
    else:
        show_end_device(enddevices.EndDevice[index])
    
@router.add('/end-devices/add')
async def add_enddevice():
    # resp = requests.get("https://127.0.0.1:7443/admin/enddevices", 
    #                     cert=('/home/os2004/tls/certs/admin.pem', '/home/os2004/tls/private/admin.pem'),
    #                     verify="/home/os2004/tls/certs/ca.pem")
    #enddevices = xml_to_dataclass(resp.text)
    #show_list(enddevices)
    add_end_device()
    
    


@router.add('/two')
async def show_two():
    ui.label('Content Two').classes('text-2xl')


@router.add('/three')
async def show_three():
    ui.label('Content Three').classes('text-4x1')


@ui.page('/')  # normal index page (eg. the entry point of the app)
@ui.page('/{_:path}')  # all other pages will be handled by the router but must be registered to also show the SPA index page
async def main():
    
    ed_tree = []
    der_tree = []
    derp_tree = []
    
    for edev_index, ed in enumerate(enddevices.EndDevice):
        ed_tree.append({'id': ed.href, 'label': f'End Device {edev_index}' })
        if ed.DERListLink:
            resp = backend_session.get(endpoint(f"/ders/{edev_index}"))
            if resp.ok:
                derlist = xml_to_dataclass(resp.text)
                for der_index, der in enumerate(derlist.DER):
                    der_tree.append({'id': der.href, 'label': f"DER {der_index}"})
    
    resp = backend_session.get(endpoint(f"/derp"))
    if resp.ok:
        derplist: m.DERProgramList = xml_to_dataclass(resp.text)
        
        for derp_index, derp in enumerate(derplist.DERProgram):
            derp_children = []
            if derp.DefaultDERControlLink:
                resp = backend_session.get(endpoint(f"/derp/{derp_index}/dderc"))
                def_ctrl: m.DefaultDERControl = xml_to_dataclass(resp.text)
                derp_children.append(dict(id=def_ctrl.href, label="Default DER Control"))
            
            #if derp.ActiveDERControlListLink:
            if derp.DERControlListLink:
                resp = backend_session.get(endpoint(f"/derp/{derp_index}/derc"))
                derp_ctl_list: m.DERControlList = xml_to_dataclass(resp.text)
                der_ctl_children = []
                for der_ctl_index, der_ctl in enumerate(derp_ctl_list.DERControl):
                    der_ctl_children.append(id=der_ctl.href, label=f"DERControl {der_ctl_index}")
                derp_children.append(dict(id=derp_ctl_list, label="DER Controls", children=der_ctl_children))
                
            derp_tree.append(dict(id=derp.href, label=f"DERP {derp_index}", children=derp_children))

    
    with ui.header(elevated=True).style('background-color: #3874c8').classes('items-center justify-between'):
        ui.label('HEADER')

    with ui.left_drawer().style('background-color: #d7e3f4'):
        
        with ui.column():
            ui.tree([
                    {'id': 'end_devices', 'label': 'End Devices', 'children': ed_tree},
                    {'id': 'der_programs', 'label': 'DER Programs', 'children': derp_tree},
                    {'id': 'usage_points', 'label': 'Usage Points'},
                    {'id': 'ders', 'label': 'DERs', 'children': der_tree},
                    {'id': 'curves', 'label': 'Curves'}
                ], label_key="label", node_key='id', on_select=lambda e: ui.notify(e.value)).classes('text-lg')
            ui.button("DER Programs", on_click=lambda: router.open(programs_list)).classes('w-64')
            ui.button("End Devices", on_click=lambda: router.open(show_end_devices)).classes('w-64')
            ui.button("Add End Device", on_click=lambda: router.open(add_enddevice)).classes('w-64')
            ui.button('One', on_click=lambda: router.open(show_one)).classes('w-64')
            ui.button('Two', on_click=lambda: router.open(show_two)).classes('w-64')
            ui.button('Three', on_click=lambda: router.open(show_three)).classes('w-64')
    
    with ui.footer().style('background-color: #3874c8'):
        ui.label('FOOTER')
    
        

    
    # this places the content which should be displayed
    router.frame().classes('w-full p-4 bg-gray-100')

ui.run()